import numpy as np
import pytest
from matplotlib import pyplot as plt
from pint import UnitRegistry

from pkdb_analysis.pk.pharmacokinetics import TimecoursePK
from pkdb_analysis.pk.pharmacokinetics_example import (
    example0,
    example1,
    example1_NoDosing,
    example2,
    example_Divoll1982_Fig1,
    example_Kim2011_Fig2,
    example_midazolam,
    show_results,
)


def test_pharmacokinetics():
    ureg = UnitRegistry()
    Q_ = ureg.Quantity
    t = np.linspace(0, 100, num=50)
    kel = 1.0
    c0 = 10.0
    dose = Q_(10.0, "mg") * Q_(1.0, "mole/g")
    c = c0 * np.exp(-kel * t)

    tcpk = TimecoursePK(
        time=Q_(t, "hr"), concentration=Q_(c, "nmol/l"), dose=dose, ureg=ureg
    )
    pk = tcpk.pk
    assert pytest.approx(pk.kel.magnitude, kel)
    assert pk.dose == Q_(0.01, "mole")
    assert pk.tmax == Q_(0.0, "hr")
    assert pk.cmax == Q_(10.0, "nmol/l")
    assert pk.vd.units == ureg.Unit("liter")


def test_pharmacokinetics_small_values():
    """Test pharmacokinetics with very small values.

    This results of replacement of the values with NaN.
    This also tests the NaN regression.
    """
    ureg = UnitRegistry()
    Q_ = ureg.Quantity
    t = np.linspace(0, 100, num=50)
    kel = 1.0
    c0 = 10.0
    dose = Q_(10.0, "mg") * Q_(1.0, "mole/g")
    c = c0 * np.exp(-kel * t)

    tcpk = TimecoursePK(
        time=Q_(t, "hr"), concentration=Q_(c, "nmol/l"), dose=dose, ureg=ureg
    )
    pk = tcpk.pk
    assert pytest.approx(pk.kel.magnitude, kel)
    assert pk.dose == Q_(0.01, "mole")
    assert pk.tmax == Q_(0.0, "hr")
    assert pk.cmax == Q_(10.0, "nmol/l")
    assert pk.vd.units == ureg.Unit("liter")


def test_mg_per_kg_units():
    # failing due to pint bug: https://github.com/hgrecco/pint/issues/1058
    # (required to go to base units first)
    ureg = UnitRegistry()
    Q_ = ureg.Quantity
    dose = Q_(10.0, "mg/kg") * Q_(1.0, "mole/g")
    print(dose)
    dose = dose.to_base_units().to_reduced_units()
    print(dose)


def test_pharmacokinetics_per_bodyweight():
    ureg = UnitRegistry()
    Q_ = ureg.Quantity
    t = np.linspace(0, 100, num=50)
    kel = 1.0
    c0 = 10.0
    dose = Q_(10.0, "mg/kg") * Q_(1.0, "mole/g")
    dose.ito("mmole/kg")  # do a hard conversion to avoid problems with units
    c = c0 * np.exp(-kel * t)

    tcpk = TimecoursePK(
        time=Q_(t, "hr"), concentration=Q_(c, "nmol/l"), dose=dose, ureg=ureg
    )
    pk = tcpk.pk
    assert pk.kel.magnitude == pytest.approx(kel)
    assert pk.dose == Q_(0.01, "mole/kg")
    assert pk.tmax == Q_(0.0, "hr")
    assert pk.cmax == Q_(10.0, "nmol/l")
    assert pk.vd.units == ureg.Unit("liter/kg")


def test_pharmacokinetics_shifted_intervention():
    ureg = UnitRegistry()
    Q_ = ureg.Quantity
    t = np.linspace(0, 100, num=50)
    kel = 1.0
    c0 = 10.0
    c = c0 * np.exp(-kel * t)
    time_shift = 10
    t = t + time_shift
    intervention_time = Q_(time_shift, "hr")
    dose = Q_(10.0, "mg") * Q_(1.0, "mole/g")

    tcpk = TimecoursePK(
        time=Q_(t, "hr"),
        concentration=Q_(c, "nmol/l"),
        dose=dose,
        ureg=ureg,
        intervention_time=intervention_time,
    )
    pk = tcpk.pk
    assert pk.kel.magnitude == pytest.approx(kel)
    assert pk.dose == Q_(0.01, "mole")
    assert pk.tmax == Q_(0.0, "hr")
    assert pk.cmax == Q_(10.0, "nmol/l")
    assert pk.vd.units == ureg.Unit("liter")


def test_pharmacokinetics_per_bodyweight2():
    ureg = UnitRegistry()
    Q_ = ureg.Quantity
    t = np.linspace(0, 100, num=50)
    kel = 1.0
    c0 = 10.0
    dose = Q_(10.0, "mg/kg")
    c = c0 * np.exp(-kel * t)

    tcpk = TimecoursePK(
        time=Q_(t, "hr"), concentration=Q_(c, "nmol/l"), dose=dose, ureg=ureg
    )
    pk = tcpk.pk
    assert pytest.approx(pk.kel.magnitude, kel)
    assert pk.dose == Q_(10.0, "mg/kg")
    assert pk.tmax == Q_(0.0, "hr")
    assert pk.cmax == Q_(10.0, "nmol/l")
    # assert pk.vd.units == ureg.Unit("liter/kg")


def test_example0():
    results = example0()
    tcpk = results[0]
    tcpk.figure()
    plt.show()

    pk = tcpk.pk
    assert pk.kel.magnitude == 1.0
    assert pk.dose == tcpk.Q_(0.01, "mole")
    assert pk.tmax == tcpk.Q_(0.0, "hr")
    assert pk.cmax == tcpk.Q_(10.0, "nmol/l")
    assert pk.vd.units == tcpk.ureg.Unit("liter")
    show_results(results)


def test_example1():
    results = example1()
    tcpk = results[0]

    pk = tcpk.pk
    assert pk.dose == tcpk.Q_(100, "mg")
    assert pk.tmax == tcpk.Q_(1.0, "hr")
    assert pk.tmaxhalf == tcpk.Q_(0.5, "hr")

    assert pk.cmax.units == tcpk.ureg.Unit("mg/l")
    assert pk.vd.units == tcpk.ureg.Unit("liter")
    show_results(results)


def test_example1_NoDosing():
    results = example1_NoDosing()
    tcpk = results[0]

    pk = tcpk.pk
    assert not hasattr(pk, "dose")
    assert not hasattr(pk, "vd")
    assert not hasattr(pk, "vdss")
    assert pk.tmax == tcpk.Q_(1.0, "hr")
    assert pk.tmaxhalf == tcpk.Q_(0.5, "hr")
    assert pk.cmax.units == tcpk.ureg.Unit("mg/l")
    show_results(results)


def test_example2():
    results = example2()
    show_results(results)


def test_example_Kim2011_Fig2():
    results = example_Kim2011_Fig2()
    show_results(results)


def test_example_Divoll1982_Fig1():
    results = example_Divoll1982_Fig1()
    pk = results[0].pk
    assert np.isnan(pk.cmaxhalf.magnitude)
    assert np.isnan(pk.tmaxhalf.magnitude)

    show_results(results)


def test_example_midazolam():
    results = example_midazolam()
    pk = results[0].pk
    print(pk)
    assert not np.isnan(pk.auc.magnitude)
    assert not np.isnan(pk.aucinf.magnitude)
    assert not np.isnan(pk.vd.magnitude)



from pkdb_analysis.data import PKData
from pkdb_analysis.test import TEST_HDF5


def _load_test_data():
    return PKData.from_hdf5(TEST_HDF5)


def test_sids():
    # check number of studies
    data = _load_test_data()
    assert len(data.study_sids) == 4


def test_intervention_filter():
    data = _load_test_data()

    def dosing_and_caffeine(d):
        return (d["measurement_type"] == "dosing") & (d["substance"] == "caffeine")

    test_data = data.filter_individual(dosing_and_caffeine)
    # check non-existing study
    assert len(test_data.study_sids) == 3

    test_data_not_concise = data.filter_individual(dosing_and_caffeine, concise=False)
    assert len(test_data_not_concise.study_sids) == 4


def test_output_filter():
    data = _load_test_data()

    def is_auc_inf(d):
        return d["measurement_type"] == "auc_inf"

    test_data = data.filter_output(is_auc_inf)
    measurement_types = test_data.outputs["measurement_type"].unique()
    assert len(measurement_types) == 1
    assert measurement_types[0] == "auc_inf"
    assert len(test_data.timecourses) > 0

    test_data = data.filter_output(is_auc_inf).delete_timecourses()
    measurement_types = test_data.outputs["measurement_type"].unique()
    assert len(measurement_types) == 1
    assert measurement_types[0] == "auc_inf"
    assert len(test_data.timecourses) == 0


def test_timecourses_filter():
    data = _load_test_data()

    def is_caffeine(d):
        return d["substance"] == "caffeine"

    test_data = data.filter_timecourse(is_caffeine)
    substances = test_data.timecourses["substance"].unique()
    assert len(substances) == 1
    assert substances[0] == "caffeine"
    assert len(test_data.timecourses) > 0

    test_data = data.filter_timecourse(is_caffeine).delete_outputs()
    substances = test_data.timecourses["substance"].unique()
    assert len(substances) == 1
    assert substances[0] == "caffeine"
    assert len(test_data.outputs) == 0


def test_subject_filter():
    data = _load_test_data()

    def smoking(d):
        return d["measurement_type"] == "smoking"

    def healthy(d):
        return d["measurement_type"] == "healthy"

    def choice_y(d):
        return d["choice"] == "Y"

    def is_healthy(d):
        return healthy(d) & choice_y(d)

    def smoker_n(d):
        return smoking(d) & (d["choice"] == "N")

    healthy_smoker_n_data = data.filter_subject([is_healthy, smoker_n])
    individual_smoking_choices = healthy_smoker_n_data.individuals[
        smoking
    ].choice.unique()
    group_smoking_choices = healthy_smoker_n_data.groups[smoking].choice.unique()

    assert "N" in individual_smoking_choices
    assert "N" in group_smoking_choices

    individual_healthy_choices = healthy_smoker_n_data.individuals[
        healthy
    ].choice.unique()
    group_healthy_choices = healthy_smoker_n_data.groups[healthy].choice.unique()

    assert "Y" in individual_healthy_choices
    assert "Y" in group_healthy_choices


def test_subject_exclude():
    data = _load_test_data()

    def smoking(d):
        return d["measurement_type"] == "smoking"

    def smoker_y(d):
        return smoking(d) & (d["choice"] == "Y")

    exclude_smoker_data = data.exclude_subject(smoker_y)

    individual_smoking_choices = exclude_smoker_data.individuals[
        smoking
    ].choice.unique()
    group_smoking_choices = exclude_smoker_data.groups[smoking].choice.unique()

    assert "Y" not in individual_smoking_choices
    assert "Y" not in group_smoking_choices


def test_group_filter_exclude():
    data = _load_test_data()

    def smoking(d):
        return d["measurement_type"] == "smoking"

    def choice_y(d):
        return d["choice"] == "Y"

    def healthy(d):
        return d["measurement_type"] == "healthy"

    def is_healthy(d):
        return healthy(d) & choice_y(d)

    def disease(d):
        return d["measurement_type"] == "disease"

    def smoker_n(d):
        return smoking(d) & (d["choice"] == "N")

    def smoker_y(d):
        return smoking(d) & choice_y(d)

    healthy_smoker_n_data = data.filter_group([is_healthy, smoker_n]).exclude_group(
        [smoker_y, disease]
    )

    group_smoking_choices = healthy_smoker_n_data.groups[smoking].choice.unique()

    assert "Y" not in group_smoking_choices
    assert "N" in group_smoking_choices

    group_healthy_choices = healthy_smoker_n_data.groups[healthy].choice.unique()

    assert "Y" in group_healthy_choices



"""
Creates test data
"""

from pkdb_analysis import PKData, PKFilter
from pkdb_analysis.query import PKDB


def load_test_studies() -> PKData:
    """Loads test studies from database.

    Ensure test studies are uploaded to database before running this script.
    """
    test_study_names = ["Test1", "Test2", "Test3", "Test4"]
    url_study_names = "__".join(test_study_names)
    pkfilter = PKFilter()
    for df_key in ["groups", "individuals", "interventions", "outputs", "data"]:
        setattr(pkfilter, df_key, {"study_name__in": url_study_names})
    return PKDB.query(pkfilter=pkfilter)


if __name__ == "__main__":
    from pkdb_analysis.test import TEST_HDF5

    pkdata = load_test_studies()
    pkdata._concise()
    pkdata.to_hdf5(TEST_HDF5)
    print(pkdata)



INDIVIDUAL_OUTPUT = {
    "value": 2.0,
    "mean": None,
    "median": None,
    "min": None,
    "max": None,
    "sd": None,
    "se": None,
    "cv": None,
    "unit": "ng",
    "value_weight": 80.0,
    "unit_weight": "kg",
    "intervention_value": 0.8,
    "intervention_unit": "ng",
}
INDIVIDUAL_NO_BODYWEIGHT_OUTPUT = {
    "value": 2.0,
    "mean": None,
    "median": None,
    "min": None,
    "max": None,
    "sd": None,
    "se": None,
    "cv": None,
    "unit": "ng",
    "value_weight": None,
    "unit_weight": None,
    "intervention_value": 0.8,
    "intervention_unit": "ng",
}
GROUP_OUTPUT = {
    "value": None,
    "mean": 4.5,
    "median": None,
    "min": None,
    "max": None,
    "sd": 4.6,
    "se": None,
    "cv": 0.4,
    "unit": "ng",
    "mean_weight": 80.0,
    "unit_weight": "kg",
    "intervention_value": 0.8,
    "intervention_unit": "ng",
}



""" caffeine specific utils"""

INTERVENTION_FIELDS = ["substance", "value", "unit", "route", "form", "application"]


def one_intervention(d):
    if len(d) == 1:
        return d


def validate_1len(d):
    assert 1 == len(d[d]), d
    return d



"""
Helper functions for body weight inferences.
Takes PKData instance and  calculates  additional outputs based on body weights of subjects.
"""
from typing import Optional, Tuple

import numpy as np
import pandas as pd
import pint
from pint import Quantity, UnitRegistry


# Define unit registry for examples
ureg = pint.UnitRegistry()

Q_ = ureg.Quantity
ureg.define("none = count")


class InferWeight(object):
    """
    a helper class for inference of results from body weights of subjects.
    """

    def __init__(self, series: pd.Series, ureg: UnitRegistry):
        self.series = series
        self.ureg = ureg
        self.weight, self.weight_field = self.get_weight()

    def per_bw(self, unit_field):
        return self.series[unit_field].endswith("/ kilogram")

    @staticmethod
    def per_bw_exp(per_bw) -> float:
        if per_bw:
            return 1
        else:
            return -1

    def get_weight(
        self,
        weight_fields=("value_weight", "mean_weight", "median_weight"),
        weight_unit_field="unit_weight",
    ) -> Tuple[Optional[Quantity], Optional[str]]:
        """ helper function to get the weight of a subject or group"""
        for weight_field in weight_fields:
            if weight_field in self.series:
                this_value = self.series[weight_field]
                this_unit = self.series[weight_unit_field]
                if this_value is not None:
                    if not np.isnan(this_value):
                        return Q_(this_value, this_unit), weight_field

        return None, None

    def bw_infer(
        self,
        unit_field="unit",
        infer_fields=("value", "mean", "median", "min", "max", "sd", "se"),
        per_bw_field="per_bw",
    ) -> pd.Series:
        """ helper function to infer values from the weight of a subject or group"""

        per_bw = self.per_bw(unit_field)
        per_bw_exp = self.per_bw_exp(per_bw)
        if self.weight is not None:
            series = self.series.copy()
            unit = self.ureg(series[unit_field])
            factor = unit * self.weight ** per_bw_exp
            for infer_field in infer_fields:
                if isinstance(series[infer_field], (float, np.ndarray)):
                    series[infer_field] = factor.m * series[infer_field]

            series[unit_field] = str(factor.u)
            series["inferred"] = True
            series[per_bw_field] = not per_bw

            return series

        return pd.Series(None, index=self.series.index)


def infer_output(series: pd.Series):
    return InferWeight(series, ureg).bw_infer()


def infer_intervention(series: pd.Series):
    return InferWeight(series, ureg).bw_infer(
        unit_field="intervention_unit",
        infer_fields=("intervention_value",),
        per_bw_field="intervention_per_bw",
    )


def infer_weight(df: pd.DataFrame, by_intervention=True, by_output=True):
    result_infer = df.dropna(subset=["unit_weight"])
    result_no_bodyweight = df[df["unit_weight"].isnull()]
    if by_output:
        result_infer_outputs = result_infer.apply(infer_output, axis="columns").dropna(
            how="all"
        )
        result_infer = result_infer.append(result_infer_outputs, ignore_index=True)
    if by_intervention:
        result_infer_outputs = result_infer.apply(
            infer_intervention, axis="columns"
        ).dropna(how="all")
        result_infer = result_infer.append(result_infer_outputs, ignore_index=True)
    result_infer = result_infer.append(result_no_bodyweight, ignore_index=True)
    return result_infer



"""
Summary tables from a PKdata instance.

Tables can be either stored to disk or uploaded to a google spreadsheet.
"""
import logging
from copy import copy
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Dict, Iterable, List, Sequence, Union

import numpy as np
import pandas as pd

from pkdb_analysis import filter, query_pkdb_data
from pkdb_analysis.data import PKData


# from gspread_pandas import Spread -> removing gspread support for now
Spread = None

logger = logging.getLogger(__name__)


def create_table_report(
    h5_data_path,
    dosing_substances: List,
    report_substances: List,
    excel_path: Path = None,
    tsv_path: Path = None,
    google_sheets: str = None,
    query_data: bool = False,
):
    """Create table report for given substance.

    h5_data_path: PKDB data in HDF5 format (via query function)
    dosing_substances: Set of substances used in Dosing, e.g. {torasemide}
    report_substances: Set of substances in reports

    excel_path: Path to excel file to write the report to
    tsv_path: Path to directory to to which the tsv are written
    google_sheets: Google sheets name for report
    query_data: boolean flag to query the data
    """
    if query_data:
        query_pkdb_data(h5_path=h5_data_path)

    # Load data
    if not h5_data_path.exists():
        raise IOError(
            f"PKDBData in HDF5 does not exist: '{h5_data_path}'. "
            f"Query the data first with the `query_data=True' flag."
        )

    pkdata = PKData.from_hdf5(h5_data_path)
    study_sids = pkdata.filter_intervention(
        f_idx=filter.f_dosing_in, substances=dosing_substances, concise=False
    ).interventions.study_sids

    pkdata = pkdata.filter_study(lambda x: x["sid"].isin(study_sids), concise=False)

    # Create table report
    table_report = TableReport(pkdata=pkdata, substances=report_substances)
    table_report.create_tables()

    # serialize table report
    if excel_path is not None:
        table_report.to_excel(excel_path)
    if tsv_path is not None:
        table_report.to_tsv(tsv_path)
    if google_sheets is not None:
        logger.error(
            f"NO SUPPORT FOR GOOGLE SHEETS: see "
            f"https://github.com/matthiaskoenig/pkdb_analysis/issues/36"
        )
        table_report.to_google_sheet(google_sheets)


@dataclass
class Parameter:
    """FIXME: Document."""

    measurement_types: Union[str, List] = "any"
    value_field: Sequence = "choice"
    substance: str = "any"
    values: Sequence = ("any",)
    only_group: bool = False
    only_individual: bool = False
    groupby: bool = True


class TableReportTypes(Enum):
    """Allowed types of table reports"""

    STUDIES = 1  # overview study content
    TIMECOURSES = 2  # overview timecourse content
    PHARMACOKINETICS = 3  # overview pharmacokinetics content


class TableReport(object):
    """Summary table for a collection of studies in PK-DB.

    Data is provided as PKData object.
    Export formats are table files or google spreadsheets.
    """

    def __init__(self, pkdata: PKData, substances: Iterable = None):
        self.pkdata = pkdata
        self.pkdata_concised = pkdata.copy()._concise()
        if substances is None:
            substances = tuple()
        self.substances = substances

        self.df_studies = None
        self.df_timecourses = None
        self.df_pharmacokinetics = None

    @staticmethod
    def _create_path(path_output):
        """Create folder."""
        if not path_output.exists():
            logger.warning(f"Path created: '{path_output.resolve()}'")
            path_output.mkdir()

    def to_excel(self, path_excel: Path):
        """Write all tables excel file."""
        self._create_path(path_excel.parent)

        sheets = {
            "studies": self.df_studies,
            "timecourses": self.df_timecourses,
            "pharmacokinetics": self.df_pharmacokinetics,
        }

        with pd.ExcelWriter(path_excel) as writer:
            for key, df in sheets.items():
                df1 = df.copy()
                # hyperlink replacements:
                df1["PKDB identifier"] = df1["PKDB identifier"].apply(
                    lambda x: f'=HYPERLINK("https://develop.pk-db.com/studies/{x}", "{x}")'
                )
                df1["PMID"] = df1["PMID"].apply(
                    lambda x: f'=HYPERLINK("https://www.ncbi.nlm.nih.gov/pubmed/{x}", "{x}")'
                )

                df1.to_excel(writer, sheet_name=key, index=False)

    def to_google_sheet(self, google_sheets: str):
        """Write all tables to google calc.

        **Google Spreadsheets**
        Set the `google_sheets` argument to use this option. The admin needs to
        create a google sheet in his google drive account.
        The name of the sheet has to be entered as the `google_sheets` argument.

        IMPORTANT:
        For google sheets to work, ask admin for the google_secret.json and
        copy it in your local config folder.
            cp  ~/.config/gspread_pandas/google_secret.json

        FOR ADMIN:
            1. Go to the Google APIs Console -> https://console.developers.google.com/
            2. Create a new project.
            3. Click Enable API and Services. Search for and enable the Google Drive API.
               Click Enable API and Services. Search for and enable the Google Sheets API.
            4. Create credentials for a Web Server to access Application Data.
            5. Name the service account and grant it a Project Role of Editor.
            6. Download the JSON file.
            7. Copy the JSON file to your code directory and rename it to google_secret.json
        """
        header_start = "A4"
        header_size = 4

        for report_type in [
            TableReportTypes.STUDIES,
            TableReportTypes.TIMECOURSES,
            TableReportTypes.PHARMACOKINETICS,
        ]:
            if report_type == TableReportTypes.STUDIES:
                table_df = self.df_studies
            elif report_type == TableReportTypes.TIMECOURSES:
                table_df = self.df_timecourses
            elif report_type == TableReportTypes.PHARMACOKINETICS:
                table_df = self.df_pharmacokinetics
                header_start = "A5"
                header_size = 5

            # hyperlink replacements:
            df = df.copy()
            df["PKDB identifier"] = df["PKDB identifier"].apply(
                lambda x: f'=HYPERLINK("https://develop.pk-db.com/studies/{x}";"{x}")'
            )
            df["PMID"] = df["PMID"].apply(
                lambda x: f'=HYPERLINK("https://www.ncbi.nlm.nih.gov/pubmed/{x}";"{x}")'
            )
            sheet_name = report_type.name.lower().capitalize()
            logger.info(f"Writing: {google_sheets}.{sheet_name}")
            spread = Spread(google_sheets)
            sheet = spread.find_sheet(sheet_name)
            sheet.resize(header_size, len(df.columns))

            spread.df_to_sheet(
                df,
                index=False,
                headers=False,
                sheet=sheet_name,
                start=header_start,
                replace=False,
            )

    def to_tsv(self, path_output: Path, suffix="tsv", sep="\t", index=None, **kwargs):
        """Write all sheets to TSV."""
        self._create_path(path_output)

        self.df_studies.to_csv(
            path_output / f"studies.{suffix}", sep=sep, index=index, **kwargs
        )
        self.df_pharmacokinetics.to_csv(
            path_output / f"pharmacokinetics.{suffix}", sep=sep, index=index, **kwargs
        )
        self.df_timecourses.to_csv(
            path_output / f"timecourses.{suffix}", sep=sep, index=index, **kwargs
        )

    def create_tables(self):
        """Creates all output tables in given output_path."""
        self.df_studies = self.create_table(report_type=TableReportTypes.STUDIES)
        self.df_timecourses = self.create_table(
            report_type=TableReportTypes.TIMECOURSES
        )
        self.df_pharmacokinetics = self.create_table(
            report_type=TableReportTypes.PHARMACOKINETICS
        )

    def create_table(self, report_type: TableReportTypes) -> pd.DataFrame:
        """Creates a summary table from PKData.

        :param: substances: iterable of PKDB substance ids for filtering of data
        """
        if not isinstance(report_type, TableReportTypes):
            raise ValueError(
                f"'report_type' must be TableReportTypes, but "
                f"is: {type(report_type), report_type}"
            )

        logger.info(f"Create TableReport: {report_type}")

        if report_type == TableReportTypes.STUDIES:
            return self.studies_table()
        elif report_type == TableReportTypes.TIMECOURSES:
            return self.timecourses_table()
        elif report_type == TableReportTypes.PHARMACOKINETICS:
            return self.pks_table()

    def studies_table(self) -> pd.DataFrame:
        """
        Changes studies in place.
        Changes study_keys in place.
        """
        table_keys = []
        table_df = self.pkdata.studies.df.copy()

        subject_info = {
            "sex": Parameter(measurement_types=["sex"], value_field=["choice"]),
            "age": Parameter(
                measurement_types=["age"], value_field=["mean", "median", "value"]
            ),
            "weight": Parameter(
                measurement_types=["weight"], value_field=["mean", "median", "value"]
            ),
            "height": Parameter(
                measurement_types=["height"], value_field=["mean", "median", "value"]
            ),
            "body mass index": Parameter(
                measurement_types=["bmi"], value_field=["mean", "median", "value"]
            ),
            "ethnicity": Parameter(
                measurement_types=["ethnicity"], value_field=["choice"]
            ),
            # pk effecting factors
            "healthy": Parameter(measurement_types=["healthy"], value_field=["choice"]),
            "medication": Parameter(
                measurement_types=["medication"], value_field=["choice"]
            ),
            "smoking": Parameter(measurement_types=["smoking"], value_field=["choice"]),
            "oral contraceptives": Parameter(
                measurement_types=["oral contraceptives"], value_field=["choice"]
            ),
            "overnight fast": Parameter(
                measurement_types=["overnight fast"], value_field=["choice"]
            ),
            "CYP1A2 genotype": Parameter(
                measurement_types=["CYP1A2 genotype"], value_field=["choice"]
            ),
            "abstinence alcohol": Parameter(
                measurement_types=["abstinence alcohol"],
                value_field=["mean", "median", "value", "min", "max"],
            ),
        }

        intervention_info = {
            "dosing amount": Parameter(
                measurement_types=["dosing", "qualitative dosing"],
                value_field=["value"],
            ),
            "dosing route": Parameter(
                measurement_types=["dosing", "qualitative dosing"],
                value_field=["route"],
            ),
            "dosing form": Parameter(
                measurement_types=["dosing", "qualitative dosing"], value_field=["form"]
            ),
        }
        outputs_info = {
            "quantification method": Parameter(
                measurement_types="any", value_field=["method"]
            ),
        }

        s_keys = list(subject_info.keys())
        i_keys = list(intervention_info.keys())
        o_keys = list(outputs_info.keys())

        table_keys.extend(["Subjects_individual", "Subjects_groups"])

        for keys in [s_keys, i_keys, o_keys]:
            table_keys.extend(copy(keys))
            keys.append("sid")

        studies_interventions = table_df.apply(
            self._add_information,
            args=(self.pkdata, intervention_info, "interventions"),
            axis=1,
        )
        studies_group = table_df.apply(
            self._add_information,
            args=(self.pkdata, self.pkdata_concised, subject_info, "groups"),
            axis=1,
        )
        studies_group[["Subjects_individual", "Subjects_groups"]] = studies_group[
            ["Subjects_individual", "Subjects_groups"]
        ].astype(int)

        studies_individuals = table_df.apply(
            self._add_information,
            args=(self.pkdata, subject_info, "individuals"),
            axis=1,
        )
        table_df = pd.merge(
            table_df,
            self._combine(
                studies_group[[*s_keys, "Subjects_individual", "Subjects_groups"]],
                studies_individuals[s_keys],
            ),
            on="sid",
        )
        table_df = pd.merge(table_df, studies_interventions[i_keys], on="sid")
        studies_outputs = table_df.apply(
            self._add_information,
            args=(self.pkdata, self.pkdata_concised, outputs_info, "outputs"),
            axis=1,
        )
        studies_timecourses = table_df.apply(
            self._add_information,
            args=(self.pkdata, self.pkdata_concised, outputs_info, "timecourses"),
            axis=1,
        )
        table_df = pd.merge(
            table_df,
            self._combine(studies_outputs[o_keys], studies_timecourses[o_keys]),
        )
        # reformating things
        table_keys, table_df = self._format_table_information(
            table_df=table_df, table_keys=table_keys
        )
        table_final_df = table_df[table_keys]
        return table_final_df

    def timecourses_table(self) -> pd.DataFrame:
        table_keys = []
        table_df = self.pkdata.studies.df.copy()

        timecourse_fields = {
            "individual": {"value_field": ["value"], "only_individual": True},
            "group": {"value_field": ["mean", "median"], "only_group": True},
            "error": {"value_field": ["sd", "se", "cv"], "only_group": True},
            "plasma/blood": {
                "value_field": ["tissue"],
                "values": ["plasma", "blood", "serum"],
                "groupby": False,
            },
            "urine": {"value_field": ["tissue"], "values": ["urine"], "groupby": False},
            "saliva": {
                "value_field": ["tissue"],
                "values": ["saliva"],
                "groupby": False,
            },
        }
        timecourse_info = {}
        for substance in self.substances:
            for key, p_kwargs in timecourse_fields.items():
                this_key = f"{substance}_timecourses_{key}"
                timecourse_info[this_key] = Parameter(
                    measurement_types="any", substance=substance, **p_kwargs
                )
                table_keys.append(this_key)

        table_df = table_df.apply(
            self._add_information,
            args=(self.pkdata, self.pkdata_concised, timecourse_info, "timecourses"),
            axis=1,
        )
        table_df = table_df.fillna(" ")

        # reformating things
        table_keys, table_df = self._format_table_information(
            table_df=table_df, table_keys=table_keys
        )
        return table_df[table_keys]

    def pks_table(self) -> pd.DataFrame:
        table_keys = []
        table_df = self.pkdata.studies.df.copy()
        pks_info = {}
        for substance in self.substances:
            pks_info_substance = {
                f"{substance}_plasma/blood": Parameter(
                    substance=f"{substance}",
                    value_field=["tissue"],
                    values=["plasma", "blood", "serum"],
                    groupby=False,
                ),
                f"{substance}_urine": Parameter(
                    substance=f"{substance}",
                    value_field=["tissue"],
                    values=["urine"],
                    groupby=False,
                ),
                f"{substance}_saliva": Parameter(
                    substance=f"{substance}",
                    value_field=["tissue"],
                    values=["saliva"],
                    groupby=False,
                ),
                f"{substance}_vd_individual": Parameter(
                    measurement_types=["vd"],
                    substance=f"{substance}",
                    value_field=["value"],
                    only_individual=True,
                ),
                f"{substance}_vd_group": Parameter(
                    measurement_types=["vd"],
                    substance=f"{substance}",
                    value_field=["mean", "median"],
                    only_group=True,
                ),
                f"{substance}_vd_error": Parameter(
                    measurement_types=["vd"],
                    substance=f"{substance}",
                    value_field=["sd", "se", "cv"],
                    only_group=True,
                ),
                f"{substance}_clearance_individual": Parameter(
                    measurement_types=["clearance"],
                    substance=f"{substance}",
                    value_field=["value"],
                    only_individual=True,
                ),
                f"{substance}_clearance_group": Parameter(
                    measurement_types=["clearance"],
                    substance=f"{substance}",
                    value_field=["mean", "median"],
                    only_group=True,
                ),
                f"{substance}_clearance_error": Parameter(
                    measurement_types=["clearance"],
                    substance=f"{substance}",
                    value_field=["sd", "se", "cv"],
                    only_group=True,
                ),
                f"{substance}_auc_individual": Parameter(
                    measurement_types=["auc_end", "auc_inf"],
                    substance=f"{substance}",
                    value_field=["value"],
                    only_individual=True,
                ),
                f"{substance}_auc_group": Parameter(
                    measurement_types=["auc_end", "auc_inf"],
                    substance=f"{substance}",
                    value_field=["mean", "median"],
                    only_group=True,
                ),
                f"{substance}_auc_error": Parameter(
                    measurement_types=["auc_end", "auc_inf"],
                    substance=f"{substance}",
                    value_field=["sd", "se", "cv"],
                    only_group=True,
                ),
                f"{substance}_thalf_individual": Parameter(
                    measurement_types=["thalf"],
                    substance=f"{substance}",
                    value_field=["value"],
                    only_individual=True,
                ),
                f"{substance}_thalf_group": Parameter(
                    measurement_types=["thalf"],
                    substance=f"{substance}",
                    value_field=["mean", "median"],
                    only_group=True,
                ),
                f"{substance}_thalf_error": Parameter(
                    measurement_types=["thalf"],
                    substance=f"{substance}",
                    value_field=["sd", "se", "cv"],
                    only_group=True,
                ),
                f"{substance}_cmax_individual": Parameter(
                    measurement_types=["cmax"],
                    substance=f"{substance}",
                    value_field=["value"],
                    only_individual=True,
                ),
                f"{substance}_cmax_group": Parameter(
                    measurement_types=["cmax"],
                    substance=f"{substance}",
                    value_field=["mean", "median"],
                    only_group=True,
                ),
                f"{substance}_cmax_error": Parameter(
                    measurement_types=["cmax"],
                    substance=f"{substance}",
                    value_field=["sd", "se", "cv"],
                    only_group=True,
                ),
                f"{substance}_tmax_individual": Parameter(
                    measurement_types=["tmax"],
                    substance=f"{substance}",
                    value_field=["value"],
                    only_individual=True,
                ),
                f"{substance}_tmax_group": Parameter(
                    measurement_types=["tmax"],
                    substance=f"{substance}",
                    value_field=["mean", "median"],
                    only_group=True,
                ),
                f"{substance}_tmax_error": Parameter(
                    measurement_types=["tmax"],
                    substance=f"{substance}",
                    value_field=["sd", "se", "cv"],
                    only_group=True,
                ),
                f"{substance}_kel_individual": Parameter(
                    measurement_types=["kel"],
                    substance=f"{substance}",
                    value_field=["value"],
                    only_individual=True,
                ),
                f"{substance}_kel_group": Parameter(
                    measurement_types=["kel"],
                    substance=f"{substance}",
                    value_field=["mean", "median"],
                    only_group=True,
                ),
                f"{substance}_kel_error": Parameter(
                    measurement_types=["kel"],
                    substance=f"{substance}",
                    value_field=["sd", "se", "cv"],
                    only_group=True,
                ),
            }
            pks_info = {**pks_info, **pks_info_substance}

        table_df = table_df.apply(
            self._add_information,
            args=(self.pkdata, self.pkdata_concised, pks_info, "outputs"),
            axis=1,
        )
        table_keys.extend(pks_info.keys())
        table_df = table_df.fillna(" ")
        # reformating things
        table_keys, table_df = self._format_table_information(
            table_df=table_df, table_keys=table_keys
        )
        return table_df[table_keys]

    @staticmethod
    def _add_information(
        study, pkdata, pkdata_concised, measurement_types: Dict, table: str
    ):
        """FIXME: document me."""

        additional_dict = {}
        # used_pkdata = pkdata.filter_study(lambda x: x["sid"] == study.sid)
        used_pkdata = pkdata_concised
        for key in ["groups", table]:
            used_pkdata[key] = [used_pkdata[key].study_sid == study.sid]

        this_table = getattr(used_pkdata, table)
        has_info_kwargs = {"df": this_table.df, "instance_id": this_table.pk}
        group_df_df = pkdata.groups.df
        study_group_df = group_df_df[group_df_df["study_sid"] == study.sid]
        all_group = study_group_df[study_group_df["group_name"] == "all"]
        subject_size = all_group.group_count.unique()[0]
        has_info_kwargs["Subjects_individual"] = subject_size
        has_info_kwargs["Subjects_groups"] = len(used_pkdata.groups.pks)

        if table == "groups":
            additional_dict["Subjects_individual"] = has_info_kwargs[
                "Subjects_individual"
            ]
            additional_dict["Subjects_groups"] = has_info_kwargs["Subjects_groups"]

        additional_dict = {
            **{
                key: TableReport._has_info(parameter=parameter, **has_info_kwargs)
                for key, parameter in measurement_types.items()
            },
            **additional_dict,
        }

        return study.append(pd.Series(additional_dict))

    @staticmethod
    def _format_table_information(table_keys, table_df):
        """Formats table information in place."""
        table_df = table_df.rename(columns={"reference_date": "publication date"})
        table_df["PKDB identifier"] = table_df["sid"].apply(lambda x: x)
        # FIXME: https://github.com/matthiaskoenig/pkdb_analysis/issues/23
        # pubmeds do not exist (sid != pmid for curated studies with PKDB identifiers)
        table_df["PMID"] = table_df["sid"].apply(lambda x: x)
        table_keys = [
            "PKDB identifier",
            "name",
            "PMID",
            "publication date",
        ] + table_keys
        table_df.sort_values(by="name", inplace=True)
        return table_keys, table_df

    @staticmethod
    def _add_group_all_count(study_df: pd.DataFrame, pkdata: PKData):
        """Adds a column with the group count of the all group for every study.
        Changes study in place.
        """
        # get all "all" groups (FIXME: unnecessary repetition)
        all_groups = pkdata.groups_core[pkdata.groups_core["group_name"] == "all"]

        group = all_groups[(all_groups["study_name"] == study_df.name)]

        # check
        assert len(group) == 1, (len(group), study_df.name, group)
        study_df["Group_all_count"] = group["group_count"]

    @staticmethod
    def _has_info(
        df: pd.DataFrame,
        instance_id: str,
        parameter: Parameter,
        Subjects_groups: int = 0,
        Subjects_individual: int = 0,
    ):
        has_info = []
        compare_length = 0
        if parameter.substance != "any":
            df = df[df["substance"] == parameter.substance]

        if parameter.only_group:
            df = df[df["individual_pk"] == -1]

            instance_id = "group_pk"
            compare_length = Subjects_groups

        if parameter.only_individual:
            df = df[df["group_pk"] == -1]
            instance_id = "individual_pk"
            compare_length = Subjects_individual

        if not parameter.groupby:
            instance_id = "study_name"
        if len(df) == 0:
            return None

        for _, instance in df.groupby(instance_id):
            if parameter.measurement_types == "any":
                specific_info = instance
            else:
                specific_info = instance[
                    instance["measurement_type"].isin(parameter.measurement_types)
                ]

            value_types = (
                specific_info[parameter.value_field].applymap(type).stack().unique()
            )

            has_array = False
            for value in value_types:
                if value is np.ndarray:
                    has_array = True
                    choices = {True}

            if not has_array:
                this_series = specific_info[parameter.value_field].max(axis=1)
                choices = set([i for i in this_series.dropna() if i is not None])

            if "any" not in parameter.values:
                choices = choices & set(parameter.values)
            if len(choices) == 0:
                has_info.append(False)
            elif "NR" in set(choices):
                has_info.append(False)
            else:
                has_info.append(True)

        if not any(has_info) or len(has_info) == 0:
            return " "

        elif all(has_info):
            if compare_length > 0:
                if len(has_info) == compare_length:
                    return "✓"
            else:
                return "✓"
        return "⅟"

    @staticmethod
    def _strict_and_logic(series):
        elements = set(series.dropna().values)
        if len(elements) == 1:
            return list(elements)[0]
        else:
            return " "

    @staticmethod
    def _and_logic(series):
        elements = set(series.dropna().values)
        if len(elements) == 1:
            return list(elements)[0]
        else:
            return "⅟"

    @staticmethod
    def _any_logic(series):
        elements = set([values for values in series.values if values is not None])
        if "✓" in elements:
            return "✓"
        else:
            return " "

    @staticmethod
    def _combine(df1, df2):
        merged = pd.merge(df1, df2, on="sid")

        for column in [c for c in df2.columns if c != "sid"]:
            keys = [f"{column}_x", f"{column}_y"]
            merged[column] = merged[keys].apply(TableReport._and_logic, axis=1)

        return merged[df1.columns]

    @staticmethod
    def _extend_study_keys(study_keys):
        return study_keys.extend(
            [
                "PKDB identifier",
                "Name",
                "PMID",
                "publication date",
                "Subjects_individual",
                "Subjects_groups",
            ]
        )

    @staticmethod
    def _clear_sheat(spread, header_size, column_length):
        spread.update_cells(
            start=(1, header_size),
            end=(1, 1),
            vals=["" for _ in range(0, column_length * 1)],
        )



import pytest

from pkdb_analysis.reports.interactive.examples import example1


# FIXME
@pytest.mark.skip("FIXME")
def test_interactive_plot1(tmp_path):
    """Test plot creation.

    You can use the tmp_path fixture which will provide a
    temporary directory unique to the test invocation,
    created in the base temporary directory.
    """
    example1.create_plots(path=tmp_path)



"""
Helper functions for pharmacokinetic analysis.

Takes concentration~time curves in plasma as input for analysis.
Pharmacokinetic parameters are than calculated and returned.
"""
import warnings
from dataclasses import dataclass

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.pyplot import Figure
from pint import Quantity, UnitRegistry
from scipy import stats


with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    Quantity([])


@dataclass
class PKParametersNoDosing:
    compound: str
    auc: Quantity
    aucinf: Quantity
    tmax: Quantity
    cmax: Quantity
    tmaxhalf: Quantity
    cmaxhalf: Quantity
    kel: Quantity
    thalf: Quantity
    slope: Quantity
    intercept: Quantity
    r_value: float
    p_value: float
    std_err: float
    max_idx: int

    @property
    def parameters(self):
        return ["auc", "aucinf", "tmax", "cmax", "tmaxhalf", "cmaxhalf", "kel", "thalf"]

    @property
    def regression_parameters(self):
        return ["slope", "intercept", "r_value", "p_value", "std_err", "max_idx"]

    def to_dict(self):
        """Convert to dictionary.

        Splits all quantities in magnitude and unit parts.
        """
        d = {
            "compound": self.compound,
        }
        for key in self.parameters:
            q = getattr(self, key)
            d[key] = q.magnitude
            d[f"{key}_unit"] = q.units

        for key in self.regression_parameters:
            d[key] = getattr(self, key)
        return d


@dataclass
class PKParameters(PKParametersNoDosing):
    dose: Quantity
    vd: Quantity
    vdss: Quantity
    cl: Quantity

    @property
    def parameters(self):
        return super().parameters + ["dose", "vd", "vdss", "cl"]


class TimecoursePKNoDosing(object):
    """ Class for calculating pharmacokinetics from timecourses without dose information """

    def __init__(
        self,
        time: Quantity,
        concentration: Quantity,
        ureg: UnitRegistry,
        substance: str = "substance",
        min_treshold=1e6,
        **kwargs,
    ):

        self._init(time, concentration, ureg, substance, min_treshold, **kwargs)
        self.pk = self._f_pk()

    def _init(
        self,
        time: Quantity,
        concentration: Quantity,
        ureg: UnitRegistry,
        substance: str = "substance",
        min_treshold=1e8,
        **kwargs,
    ):
        self.ureg = ureg
        self.Q_ = ureg.Quantity

        if not isinstance(time, Quantity):
            raise ValueError(f"'time' must be a pint Quantity: {type(time)}")
        if not isinstance(concentration, Quantity):
            raise ValueError(
                f"'concentration' must be a pint Quantity: {type(concentration)}"
            )

        # check concentration is fitting to time
        assert time.size == concentration.size

        # for numerical simulations problems in calculations can arise
        # if values are getting too small
        cmin = np.nanmin(
            concentration[np.nonzero(concentration)]
        )  # only take non-zero values
        cmax = np.nanmax(concentration)
        if (min_treshold * cmin) < cmax:
            warnings.warn(f"Very small concentrations values are set to NaN.")
            concentration[concentration * min_treshold < cmax] = np.nan

        self.t = time
        self.c = concentration
        self.substance = substance

    def _f_pk(self) -> PKParametersNoDosing:
        """Calculates all pk parameters from given time course.

        The returned data structure can be used to
        - create a report with pk_report
        - create a visualization with pk_figure

        """
        c = self.c
        t = self.t
        # simple pk
        auc = self._auc(t, c)
        tmax, cmax = self._max(t, c)
        tmaxhalf, cmaxhalf = self._max_half(t, c)
        [slope, intercept, r_value, p_value, std_err, max_idx] = self._ols_regression(
            t, c
        )
        kel = self._kel(slope=slope)
        thalf = self._thalf(kel=kel)
        aucinf = self._aucinf(t, c, slope=slope)

        return PKParametersNoDosing(
            compound=self.substance,
            auc=auc.to_reduced_units(),
            aucinf=aucinf.to_reduced_units(),
            tmax=tmax.to_reduced_units(),
            cmax=cmax.to_reduced_units(),
            tmaxhalf=tmaxhalf.to_reduced_units(),
            cmaxhalf=cmaxhalf.to_reduced_units(),
            kel=kel.to_reduced_units(),
            thalf=thalf.to_reduced_units(),
            slope=slope.to_reduced_units(),
            intercept=intercept.to_reduced_units(),
            r_value=r_value,
            p_value=p_value,
            std_err=std_err,
            max_idx=max_idx,
        )

    def _ols_regression(self, t, c):
        """Linear regression on the log timecourse after maximal value.

        No check is performed if already in equilibrium distribution !.
        The linear regression is calculated from all data points after the maximum.

        :return:
        """
        slope_units = self.ureg.Unit(f"1/{t.units}")
        intercept_units = self.ureg.Unit(c.units)

        max_index = np.nanargmax(c)
        # at least three data points after maximum are required for a regression

        # FIXME:
        if max_index > (len(c) - 4):
            warnings.warn(
                "Regression could not be calculated, "
                "at least 3 data points after maximum required."
            )
            return [self.Q_(np.nan, slope_units), self.Q_(np.nan, intercept_units)] + [
                np.nan
            ] * 4

        # linear regression start regression on data point after maximum
        x = t.magnitude[max_index + 1 :]
        y = np.log(c.magnitude[max_index + 1 :])

        # using mask to remove nan values
        mask = ~np.isnan(x) & ~np.isnan(y)
        slope, intercept, r_value, p_value, std_err = stats.linregress(x[mask], y[mask])

        # handle possible regression issues
        if np.isnan(slope) or np.isnan(intercept):
            warnings.warn("Regression could not be calculated on timecourse curve.")
        elif slope > 0.0:
            warnings.warn(
                "Regression gave a positive slope, "
                "resulting in a negative elimination rate. "
                "Slope is set to NaN."
            )
            slope = np.nan
            intercept = np.nan

        slope = self.Q_(slope, slope_units)
        intercept = self.Q_(intercept, intercept_units)

        return [slope, intercept, r_value, p_value, std_err, max_index]

    def _auc(self, t, c):
        """ Calculates the area under the curve (AUC) via trapezoid rule """
        auc = np.sum((t[1:] - t[0:-1]) * (c[1:] + c[0:-1]) / 2.0)
        return auc

    def _aucinf(self, t, c, slope=None):
        """Calculates the area under the curve (AUC) via trapezoid rule
        and extrapolated to infinity"""
        auc = self._auc(t, c)

        # by integrating from tend to infinity for c[-1]exp(slope * t) we get
        auc_d = -c[-1] / slope

        if auc_d > auc:
            warnings.warn(f"AUC(t-oo) > AUC(0-tend), no AUC(0-oo) calculated.")
            # return self.Q_(np.nan, auc.units)

        if auc_d > 0.25 * auc:
            # If the % extrapolated is greater than 20%, than the total AUC may be unreliable.
            # The unreliability of the data is not due to a calculation error. Instead it
            # indicates that more sampling is needed for an accurate estimate of the elimination
            # rate constant and the observed area under the curve.
            warnings.warn(
                f"AUC(t-oo) is >25% ({round((auc_d/auc*100).magnitude, 2)}%) of total AUC, "
                f"calculation may be unreliable."
            )

        return auc + auc_d

    def _max(self, t, c):
        """Returns timepoint of maximal value and maximal value based on curve.

        The tmax depends on the value of both the absorption rate constant (ka)
        and the elimination rate constant (kel).

        :return: tuple (tmax, cmax)
        """
        idx = np.nanargmax(c)
        return t[idx], c[idx]

    def _max_half(self, t, c):
        """Calculates timepoint of half maximal value.

        The max half is the timepoint before reaching the maximal value.

        :return: tuple (tmax, cmax)
        """
        try:
            idx = np.nanargmax(c)
            if idx == len(c) - 1:
                warnings.warn("No MAXIMUM reached within time course, last value used.")
            if idx == 0:
                # no maximum in time course
                return self.Q_(np.nan, t.units), self.Q_(np.nan, c.units)

            cmax = c[idx]
            tnew = t[:idx]
            cnew = np.abs(c[:idx] - 0.5 * cmax)
            idx_half = np.nanargmin(cnew)
            return tnew[idx_half], c[idx_half]
        except ValueError:
            # often only NaN values before maximum (e.g., iv dosing)
            return self.Q_(np.nan, t.units), self.Q_(np.nan, c.units)

    def _kel(self, slope):
        """
        Elimination half-life (t1/2) and elimination rate constant were
        computed by linear regression of the log plasma concentrations vs.
        time after the maximum.

        c(t) = c0 * exp(-kel*t)
        log(c(t)) = log(c0) - kel* t

        Elimination rate constant (kel): Fractional rate of drug removal from the body.
        This rate is constant in first-order kinetics and is independent of
        drug concentration in the body. kel is the slope of the plasma concentration-time
        line (on a logarithmic y scale).
        """
        return -slope

    def _kel_cv(self, std_err=None, slope=None):
        return std_err / slope

    def _thalf(self, kel):
        """Calculates the half-life using the elimination constant.

        Definition: Time it takes for the plasma concentration or the amount
        of drug in the body to be reduced by 50%.

        Half-life determines the length of the drug effect. It also indicates
        whether accumulation of the drug will occur under a multiple dosage regimen
        and it is essential to decide on the appropriate dosing interval.

        If no kel is provided t and c must be provided for calculation
        of the elimination constant.
        """
        # np.log is natural logarithm, i.e. ln(2)
        return np.log(2) / kel

    def _thalf_cv(self, kel_cv):
        return np.log(2) / kel_cv

    def info(self) -> str:
        """Information string for given pharmacokinetic information.

        :return:
        """
        lines = []
        lines.append("-" * 80)
        lines.append(self.pk.compound)
        lines.append("-" * 80)
        for key in self.pk.regression_parameters:
            lines.append("{:<12}: {:>3.3f}".format(key, getattr(self.pk, key)))
        lines.append("-" * 80)
        for key in self.pk.parameters:
            lines.append("{:<12}: {:P}".format(key, getattr(self.pk, key)))
        lines.append("-" * 80)
        return "\n".join(lines)

    def figure(self, title=None) -> Figure:
        """
        Create figure from time course and pharmacokinetic parameters.
        """
        fig, (ax1, ax2) = plt.subplots(nrows=1, ncols=2, figsize=(10, 5))
        for ax in (ax1, ax2):
            if title is None:
                title = self.pk.compound
            ax.set_title(title)
            ax.set_xlabel(f"time [{self.t.units}]")
            ax.set_ylabel(f"concentration [{self.c.units}]")

        t = self.t.magnitude
        c = self.c.magnitude
        slope = self.pk.slope.magnitude
        intercept = self.pk.intercept.magnitude
        max_idx = self.pk.max_idx
        if max_idx is None or np.isnan(max_idx):
            max_idx = c.size - 1

        # pharmacokinetic parameters
        cmax = self.pk.cmax.magnitude
        tmax = self.pk.tmax.magnitude

        # aucinf
        tend = t[-1]
        cend = c[-1]

        kwargs = {
            "linestyle": "-",
            "linewidth": 2.0,
        }
        for ax in (ax1, ax2):

            # auc
            ax.fill_between(
                t, np.zeros_like(c), c, color="green", alpha=0.2, label="AUCend"
            )
            ax.plot((tend, tend), (0, cend), linestyle="-", color="black")

            # aucinf
            t_aucinf = np.linspace(0, 0.3 * tend, 50)  # interpolate by 30%
            c_aucinf = cend * np.exp(slope * t_aucinf)
            ax.fill_between(
                tend + t_aucinf,
                c_aucinf,
                np.zeros_like(c_aucinf),
                color="red",
                alpha=0.2,
                label="AUCinf",
            )
            ax.plot(tend + t_aucinf, c_aucinf, linestyle="-", color="black")

            # fit
            if not np.isnan(slope):
                ax.plot(
                    t,
                    np.exp(intercept) * np.exp(slope * t),
                    "-",
                    color="blue",
                    label="fit",
                    linewidth=2.0,
                )
            # cmax (hline)
            ax.plot((0, tmax), (cmax, cmax), linestyle="--", color="black")
            # tmax (vline)
            ax.plot((tmax, tmax), (0, cmax), linestyle="--", color="black")

            ax.plot(t, c, "o-", color="black", markersize=6, linewidth=2)

            # plot data points used for fitting
            if not np.isnan(slope):
                if max_idx < c.size - 1:
                    ax.plot(
                        t[max_idx + 2 :],
                        c[max_idx + 2 :],
                        "s",
                        color="blue",
                        linewidth=2,
                        markersize=8,
                    )
            ax.annotate("(tmax, cmax)", xy=(tmax, cmax))

        ax1.set_ylim(bottom=0)
        ax2.set_yscale("log")

        for ax in (ax1, ax2):
            ax.set_xlim(left=0)
            ax.legend()
        return fig


class TimecoursePK(TimecoursePKNoDosing):
    """ Class for calculating pharmacokinetics from timecourses. """

    def __init__(
        self,
        time: Quantity,
        concentration: Quantity,
        dose: Quantity,
        ureg: UnitRegistry,
        intervention_time: Quantity = None,
        substance: str = "substance",
        min_treshold=1e6,
        **kwargs,
    ):
        """Pharmacokinetics parameters are calculated for a single dose experiment.

        TODO: support errors on concentrations which are then used in calculation
        FIXME: ctype is used in kwargs for "value", "mean", "median", but not
         processed

        tmax values are reported relative to intervention time

        :param time: ndarray (with units)
        :param concentration: ndarray (with units)
        :param dose: dose of the test substance (with units)
        :param ureg: unit registry, allowing to calculate the pk in the respective unit system
        :param substance: name of compound/substance
        :param intervention_time: time of intervention (with unit)

        :return: pharmacokinetic parameters
        """
        self._init(time, concentration, ureg, substance, min_treshold, **kwargs)
        if intervention_time is None:
            intervention_time = self.Q_(0.0, "hr")
        if dose is None:
            dose = self.Q_(np.nan, "mg")
        if not isinstance(dose, Quantity):
            raise ValueError(f"'dose' must be a pint Quantity: {type(dose)}")
        if not isinstance(intervention_time, Quantity):
            raise ValueError(
                f"'intervention_time' must be a pint Quantity: {type(intervention_time)}"
            )

        # check dimensionality of dose
        dr = (
            dose.to_base_units().to_reduced_units()
        )  # see https://github.com/hgrecco/pint/issues/1058
        if not (
            dr.check("[mass]")
            or dr.check("[substance]")
            or dr.check("[mass]/[mass]")
            or dr.check("[substance]/[mass]")
        ):
            warnings.warn(
                f"dose_reduced.dimensionality must either be in '[mass]', '[substance']', '[mass]/[mass]' or '[substance]/[mass]'"
                f"The given units are: '{dr.dimensionality}' for {dr.units}. "
                f"Check that dose units are correct."
            )
            raise ValueError(
                f"Incorrect dimensionality '{dr.dimensionality}' for dose: {dose.units}"
            )

        self.dose = dose
        # convert dosing time to units of timecourse
        self.intervention_time = intervention_time.to(time.units)
        self.substance = substance
        self.pk = self._f_pk()

    def _f_pk(self) -> PKParameters:
        """Calculates all pk parameters from given time course.

        The returned data structure can be used to
        - create a report with pk_report
        - create a visualization with pk_figure

        """
        # calculate all results relative to the intervention time
        t = self.t - self.intervention_time
        c = self.c

        # simple pk
        auc = self._auc(t, c)
        tmax, cmax = self._max(t, c)
        tmaxhalf, cmaxhalf = self._max_half(t, c)

        [slope, intercept, r_value, p_value, std_err, max_idx] = self._ols_regression(
            t, c
        )

        kel = self._kel(slope=slope)
        thalf = self._thalf(kel=kel)
        aucinf = self._aucinf(t, c, slope=slope)

        if self.dose is not None and not np.isnan(self.dose.magnitude):
            # parameters depending on dose
            vdss = self._vdss(dose=self.dose, intercept=intercept)
            vd = self._vd(aucinf=aucinf, dose=self.dose, kel=kel)
            cl = kel * vd
        else:
            vd_units = self.dose.units / (auc.units / kel.units)
            vdss = self.Q_(np.nan, vd_units)
            vd = self.Q_(np.nan, vd_units)
            cl = self.Q_(np.nan, kel.units * vd.units)

        # perform unit normalization on volumes
        vd.to_base_units().to_reduced_units(),  # see https://github.com/hgrecco/pint/issues/1058
        vdss.to_base_units().to_reduced_units(),  # see https://github.com/hgrecco/pint/issues/1058
        for vd_par in [vd, vdss]:
            if vd_par.check("[length] ** 3"):
                vd_par.ito("liter")
            elif vd_par.check("[length] ** 3/[mass]"):
                vd_par.ito("liter/kg")

        return PKParameters(
            compound=self.substance,
            dose=self.dose.to_reduced_units(),
            auc=auc.to_reduced_units(),
            aucinf=aucinf.to_reduced_units(),
            tmax=tmax.to_reduced_units(),
            cmax=cmax.to_reduced_units(),
            tmaxhalf=tmaxhalf.to_reduced_units(),
            cmaxhalf=cmaxhalf.to_reduced_units(),
            kel=kel.to_reduced_units(),
            thalf=thalf.to_reduced_units(),
            vd=vd,
            vdss=vdss,
            cl=cl.to_reduced_units(),
            slope=slope.to_reduced_units(),
            intercept=intercept.to_reduced_units(),
            r_value=r_value,
            p_value=p_value,
            std_err=std_err,
            max_idx=max_idx,
        )

    def _vdss(self, dose, intercept=None):
        """
        Apparent volume of distribution.
        Not a physical space, but a dilution space.

        Definition: Fluid volume that would be required to contain the amount of drug present
        in the body at the same concentration as in the plasma.

        Calculation: The Vd is calculated as the ratio of the dose present in the body
        and its plasma concentration, when the distribution of the drug between
        the tissues and the plasma is at equilibrium. The extrapolated plasma concentration
        at time 0, C(0), is back-extrapolated from the slope of the elimination phase of
        the semilogarithmic plasma concentration vs. time decay curve.

        If both the dose (mg/kg) and the drug concentration in plasma
        (the Y intercept of the terminal component of the plasma drug concentration [PDC] versus time curve)
        are known, then an "apparent" volume of distribution can be calculated from Vd = dose/PDC.
        This theoretical volume describes the volume to which the drug must be distributed
        if the concentration in plasma represents the concentration throughout the body
        (i.e., distribution has reached equilibrium).
        The term "apparent" underscores the fact that where the drug is distributed cannot be
        determined from Vd; only that it goes somewhere.
        """
        return dose / self.Q_(np.exp(intercept.magnitude), intercept.units)

    def _vd(self, aucinf, dose, kel):
        """
        Apparent volume of distribution.
        Not a physical space, but a dilution space.

        Volume of distribution is calculated via
            vd = Dose/(AUC_inf*kel)
        """
        vd = dose / (aucinf * kel)
        return vd



import os
from collections import namedtuple

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.lines import Line2D
from matplotlib.ticker import FormatStrFormatter

from pkdb_analysis.deprecated.utils import abs_idx, group_idx, individual_idx, rel_idx


plt.style.use("seaborn-white")
import matplotlib.ticker as ticker
import pint


ureg = pint.UnitRegistry()

plt.rcParams.update(
    {
        "axes.labelsize": "20",
        "axes.labelweight": "bold",
        "axes.titlesize": "medium",
        "axes.titleweight": "bold",
        "legend.fontsize": "20",
        "xtick.labelsize": "20",
        "ytick.labelsize": "20",
        "figure.facecolor": "1.00",
    }
)

import matplotlib.font_manager as font_manager


font = font_manager.FontProperties(
    family="Roboto Mono",
    weight="normal",
    style="normal",
    size=16,
)


class FigureTemplate(object):
    def __init__(self, data_idx, intervention_type, output_type):
        self.data_idx = (data_idx,)
        self.intervention_type = intervention_type
        self.output_type = output_type
        self.figure = (False,)
        self.ax = False

    def data_subset(self, data):
        return data.loc[self.data_idx]

    def create_figure(self, data):
        if len(self.data_subset(data)) > 0:
            self.figure, self.ax = plt.subplots(nrows=1, ncols=1, figsize=(15, 15))


PlotCategory = namedtuple("PlotCategory", ["name", "color", "marker", "data_idx"])


def create_figures(df_data):

    figure_templates = [
        FigureTemplate(
            data_idx=abs_idx(df_data, "unit_intervention") & abs_idx(df_data, "unit"),
            intervention_type="abs",
            output_type="abs",
        ),
        FigureTemplate(
            data_idx=rel_idx(df_data, "unit_intervention") & abs_idx(df_data, "unit"),
            intervention_type="rel",
            output_type="abs",
        ),
        FigureTemplate(
            data_idx=abs_idx(df_data, "unit_intervention") & rel_idx(df_data, "unit"),
            intervention_type="abs",
            output_type="rel",
        ),
        FigureTemplate(
            data_idx=rel_idx(df_data, "unit_intervention") & rel_idx(df_data, "unit"),
            intervention_type="rel",
            output_type="rel",
        ),
    ]
    figure, axes = plt.subplots(nrows=2, ncols=2, figsize=(30, 30))
    axes
    for n, figure_template in enumerate(figure_templates):
        figure_template.figure = figure
        figure_template.ax = axes.flatten()[n]
        # figure_template.create_figure(df_data)

    return figure_templates


pk_name = {
    "auc_end": "$AUC_{end}$",
    "auc_inf": "$AUC_{\infty}$",
    "clearance": "$Clearance$",
    "cmax": "$C_{max}$",
    "kel": "$k_{el}$",
    "thalf": "$t_{half}$",
    "tmax": "$t_{max}$",
    "vd": "$Vd$",
}


def create_plots(df_data, categories, fig_path, log_y=False):
    measurement_type = df_data["measurement_type"].unique()[0]
    substance = df_data["substance"].unique()[0]
    figures = create_figures(df_data)

    for figure in figures:
        if figure.ax:
            df_figure = figure.data_subset(df_data)
            max_values = np.array([df_figure["value"].max(), df_figure["mean"].max()])
            max_values = max_values[~np.isnan(max_values)]
            df_figure_max = np.max(max_values) * 1.05

            df_figure_x_max = df_figure["value_intervention"].max() * 1.05

            df_figure_min = (
                0  # min([df_figure["value"].min(), df_figure["mean"].min()]) / 1.05
            )

            df_individual = df_figure[individual_idx(df_figure)]
            df_group = df_figure[group_idx(df_figure)]

            legend_elements = []

            for plot_category in categories:

                df_category = df_individual[plot_category.data_idx(df_individual)]

                units = df_category["unit"].unique()
                units_intervention = df_category["unit_intervention"].unique()
                assert len(units) <= 1, units
                try:
                    unit = units[0]
                    u_unit = ureg(unit)
                    unit_intervention = units_intervention[0]
                    u_unit_intervention = ureg(unit_intervention)

                    y_axis_label = (
                        f"{substance} {pk_name.get(measurement_type, measurement_type)}"
                    )

                    figure.ax.set_ylabel(f"{y_axis_label} [{u_unit.u :~P}]")
                    substance_internvetion = df_category[
                        "substance_intervention"
                    ].unique()[0]
                    x_label = "$Dose_{" + substance_internvetion + "}$"
                    figure.ax.set_xlabel(f"{x_label} [{u_unit_intervention.u :~P}]")
                    figure.ax.yaxis.set_major_formatter(FormatStrFormatter("%.2f"))

                except IndexError:
                    pass

                individuals_number = len(df_category)

                for (calculated, inferred), df_individuals in df_category.groupby(
                    ["calculated", "inferred"]
                ):

                    if calculated:
                        marker = "s"
                    else:
                        marker = plot_category.marker

                    x = df_individuals["value_intervention"]
                    y = df_individuals["value"]
                    figure.ax.scatter(
                        x,
                        y,
                        color=plot_category.color,
                        marker=marker,
                        alpha=0.7,
                        label=None,
                        s=20,
                    )

                df_category = df_group[plot_category.data_idx(df_group)]
                df_category = df_category[df_category["mean"].notnull()]
                group_number = len(df_category)
                total_group_individuals = 0
                x_group_max = df_category["value_intervention"].max()
                y_group_max = df_category["mean"].max()

                for i, df_row in df_category.iterrows():
                    if df_row["inferred"]:
                        marker = "v"

                    elif df_row["calculated"]:
                        marker = "s"
                    else:
                        marker = plot_category.marker

                    x_group = df_row["value_intervention"]
                    y_group = df_row["mean"]

                    if figure.output_type == "rel":
                        xerr_group = (
                            df_row[("weight", "sd")] / df_row[("weight", "mean")]
                        ) * x_group
                    else:
                        xerr_group = 0
                    yerr_group = df_row["se"]
                    group_count = df_row["group_count"]
                    total_group_individuals += group_count

                    figure.ax.errorbar(
                        x_group,
                        y_group,
                        yerr=yerr_group,
                        xerr=xerr_group,
                        color=plot_category.color,
                        fmt=marker,
                        ms=group_count + 5,
                        alpha=0.7,
                    )
                    # for i, txt in enumerate(df_category[('study', '')]):
                    txt = df_row[("study", "")]
                    figure.ax.annotate(
                        txt,
                        (
                            x_group + (0.01 * x_group_max),
                            y_group + (0.01 * y_group_max),
                        ),
                        alpha=0.7,
                    )

                label_text = f"{plot_category.name:<10} I: {individuals_number:<3} G: {group_number:<2} TI: {int(total_group_individuals + individuals_number):<3}"
                print(label_text)
                # "
                label = Line2D(
                    [0],
                    [0],
                    marker="o",
                    color="w",
                    label=label_text,
                    markerfacecolor=plot_category.color,
                    markersize=10,
                )
                legend_elements.append(label)

            legend2_elements = [
                Line2D(
                    [0],
                    [0],
                    marker="o",
                    color="w",
                    label="PK DATA",
                    markerfacecolor="black",
                    markersize=10,
                ),
                Line2D(
                    [0],
                    [0],
                    marker="s",
                    color="w",
                    label="PK FROM TIME COURSE",
                    markerfacecolor="black",
                    markersize=10,
                ),
                Line2D(
                    [0],
                    [0],
                    marker="v",
                    color="w",
                    label="PK FROM BODY WEIGHT",
                    markerfacecolor="black",
                    markersize=10,
                ),
            ]

            legend3_elements = [
                Line2D(
                    [0],
                    [10],
                    marker="o",
                    color="w",
                    label="1",
                    markerfacecolor="black",
                    markersize=5 + 1,
                ),
                Line2D(
                    [0],
                    [20],
                    marker="o",
                    color="w",
                    label="10",
                    markerfacecolor="black",
                    markersize=5 + 10,
                ),
                Line2D(
                    [0],
                    [50],
                    marker="o",
                    color="w",
                    label="30",
                    markerfacecolor="black",
                    markersize=5 + 30,
                ),
            ]

            figure.ax.set_title(
                f"{figure.output_type}-vs-dosing_{figure.intervention_type}"
            )
            figure.ax.set_xlim(left=0, right=df_figure_x_max)
            # figure.ax.yaxis.set_major_locator(MaxNLocator(integer=True))
            figure.ax.yaxis.set_major_formatter(
                ticker.ScalarFormatter(useMathText=True)
            )

            leg1 = figure.ax.legend(
                handles=legend_elements, prop=font, loc="upper right"
            )

            leg2 = figure.ax.legend(
                handles=legend2_elements, prop=font, loc="upper left"
            )
            leg3 = figure.ax.legend(
                handles=legend3_elements,
                prop=font,
                labelspacing=1.3,
                loc=("center right"),
            )
            leg3.set_title(title="GROUP SIZE", prop=font)
            figure.ax.add_artist(leg2)
            figure.ax.add_artist(leg1)

            if log_y:
                figure.ax.set_yscale("log")
                figure.ax.set_ylim(bottom=df_figure_min, top=df_figure_max)
            else:
                figure.ax.set_ylim(bottom=0, top=df_figure_max)

    figures[0].figure.savefig(
        os.path.join(fig_path, f"{substance}_{measurement_type}.png"),
        bbox_inches="tight",
        dpi=72,
    )



"""Example to create interactive pharmacokinetics plot."""

from pathlib import Path

import pandas as pd

from pkdb_analysis import PKData
from pkdb_analysis.filter import f_effective_n_oc, f_n_oc, f_n_smoking, f_oc, f_smoking
from pkdb_analysis.reports.interactive.interactive import LegendArgs as LA
from pkdb_analysis.reports.interactive.interactive import PlottingParameter as PP
from pkdb_analysis.reports.interactive.interactive import interactive_plot_factory
from pkdb_analysis.test import TEST_HDF5


# substances used in the interventions
INTERVENTION_SUBSTANCES = {"caffeine"}

# substances measured in the outputs
OUTPUT_SUBSTANCES = {"caffeine"}

# create plotting categories
PLOTTING_CATEGORIES = [
    PP(["auc_end"], []),
    PP(["auc_inf"], []),
    PP(["vd"], []),
    PP(["thalf"], [], infer_by_output=False),
    PP(["clearance"], ["milliliter / meter ** 2 / minute"]),
    PP(["cmax"], [], infer_by_output=False),
    PP(["kel"], []),
    PP(["tmax"], [], infer_by_output=False),
    PP(["vd"], []),
]

# Change formating on specific values in specific columns
REPLACEMENTS = {
    "sex": {"NR": "not reported"},
    "intervention_route": {"iv": "intravenous"},
}

# The url which will be used to href on click.
URL = "https://develop.pk-db.com"

# The name and key of the column which has categories. The categories can be clicked on an interactive legend.
# The categories cannot be color coded.
MULTI_LEGEND = {"Study": "study_name"}
# The name and key of the column which has categories. The categories can be clicked on an interactive legend.
# The categories can be color coded.

MULTI_COLOR_LEGEND = {
    "Data type": LA("data_type"),
    "Outlier": LA(
        "outlier",
        [
            False,
        ],
    ),
    "Sex": LA("sex"),
    "Healthy": LA(
        "healthy",
        [True],
    ),
    "Lifestyle": LA(
        "life_style",
    ),
    "Administration route": LA("intervention_route"),
    "Number of interventions": LA(
        "intervention_number",
        [
            1,
        ],
    ),
    "Tissue": LA("tissue"),
}
# Information shown on hover.
TOOLTIP = [
    "study_sid:N",
    "study_name:N",
    "url:N",
    "output_pk:N",
    "group_name:N",
    "individual_name:N",
    "intervention_value:Q",
    "y:Q",
    "intervention_names:N",
    "intervention_number:Q",
    "weight:Q",
    "unit_weight:N",
    "sex:N",
]


def outlier_studies(df):
    outlier_study_names = ["Balogh1992", "Harder1988", "Harder1988", "Stille1987"]
    if df.study_name in outlier_study_names:
        return True
    else:
        return False


def lifestyle(df):
    if not df.extra[f_smoking].empty and df.extra[f_n_smoking].empty:
        return "smoking"
    elif not df.extra[f_oc].empty and df.extra[f_n_oc].empty:
        return "oral contraceptive"
    elif not df.extra[f_n_smoking].empty and not df.extra[f_effective_n_oc].empty:
        return "control"
    else:
        return "unknown"


def intervention_names(df):
    """calculates a column with a information on all applied intervention names.
    This is information is displayed on hover."""

    if isinstance(df.intervention_extra, pd.DataFrame):
        return " | ".join(df.intervention_extra["name"])
    else:
        print(df.intervention_extra)


# column name and function on how the column name is calculated.
# The column names can be used e.g. as colored multi legends.
ADDITIONAL_INFORMATION = {
    "intervention_names": intervention_names,
    "outlier": outlier_studies,
    "life_style": lifestyle,
}


def create_plots(path):
    pkdata = PKData.from_hdf5(TEST_HDF5)

    kwargs = {
        "plotting_categories": PLOTTING_CATEGORIES,
        "intervention_substances": INTERVENTION_SUBSTANCES,
        "output_substances": OUTPUT_SUBSTANCES,
        "exclude_study_names": [],
        "additional_information": ADDITIONAL_INFORMATION,
        "multi_color_legend": MULTI_COLOR_LEGEND,
        "tooltip": TOOLTIP,
        "multi_legend": MULTI_LEGEND,
        "url": URL,
        "create_json": True,
        "replacements": REPLACEMENTS,
    }

    interactive_plot_factory(pkdata, path=path, **kwargs)


if __name__ == "__main__":
    # !Delete the output after running!
    output_path = Path(__file__).parent / "results"
    # if not output_path.exists():
    #    output_path.mkdir()
    create_plots(path=output_path)



"""
This examples shows how based on time and concentration vector the
pharmacokinetic parameters can be calculated.
"""
import warnings
from pathlib import Path
from typing import Dict, List

import numpy as np
import pandas as pd
import pint
from matplotlib import pyplot as plt
from pint import Quantity

from pkdb_analysis.pk.pharmacokinetics import TimecoursePK, TimecoursePKNoDosing
from pkdb_analysis.test import TESTDATA_PATH


with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    Quantity([])

# Define unit registry for examples
ureg = pint.UnitRegistry()
Q_ = ureg.Quantity


def example0() -> List[TimecoursePK]:
    """Calculate pharmacokinetics from simulated data."""
    t = np.linspace(0, 3, num=50)
    kel = 1.0
    c0 = 10.0
    dose = Q_(10.0, "mg") * Q_(1.0, "mole/g")
    c = c0 * np.exp(-kel * t)

    tcpk = TimecoursePK(
        time=Q_(t, "hr"), concentration=Q_(c, "nmol/l"), dose=dose, ureg=ureg
    )

    return [tcpk]


def example1() -> List[TimecoursePK]:
    """Example for pharmacokinetics calculation.

    :return:
    """
    results = []
    df = pd.read_csv(
        TESTDATA_PATH / "pk" / "data_example1.csv", sep="\t", na_values="NA"
    )

    # ------------------------------------------
    # Pharmacokinetic parameter for caffeine
    # ------------------------------------------
    # get caffeine data
    dose = Q_(100, "mg")
    substance = "caffeine"

    for tissue in df.tissue.unique():
        for group in df.group.unique():
            print("substance: {}, dose: {}".format(substance, dose))
            data = df[(df.tissue == tissue) & (df.group == group)]

            # calculate pharmacokinetic information
            t = Q_(data.time.values, "hr")
            c = Q_(data.caf.values, "mg/l")
            tcpk = TimecoursePK(
                time=t, concentration=c, substance=substance, dose=dose, ureg=ureg
            )
            results.append(tcpk)

    return results


def example1_NoDosing() -> List[TimecoursePKNoDosing]:
    """Example for pharmacokinetics calculation.

    :return:
    """
    results = []
    df = pd.read_csv(
        TESTDATA_PATH / "pk" / "data_example1.csv", sep="\t", na_values="NA"
    )

    # ------------------------------------------
    # Pharmacokinetic parameter for caffeine
    # ------------------------------------------
    # get caffeine data
    substance = "caffeine"

    for tissue in df.tissue.unique():
        for group in df.group.unique():
            print("substance: {}".format(substance))
            data = df[(df.tissue == tissue) & (df.group == group)]

            # calculate pharmacokinetic information
            t = Q_(data.time.values, "hr")
            c = Q_(data.caf.values, "mg/l")
            tcpk = TimecoursePKNoDosing(
                time=t, concentration=c, substance=substance, ureg=ureg
            )
            results.append(tcpk)

    return results


def example2() -> List[TimecoursePK]:
    """Example for pharmacokinetics calculation.

    This demonstrates the extreme examples of time courses with only limited data points.
    On caffeine the regressions can be calculated, on paraxanthine this is not possible any more (only a single datapoint
    or one data point after the maximum.

    :return:
    """
    results = []
    df = pd.read_csv(
        TESTDATA_PATH / "pk" / "data_example2.csv", sep="\t", na_values="NA"
    )

    # ------------------------------------------
    # Pharmacokinetic parameter for caffeine
    # ------------------------------------------
    # get caffeine data
    for (substance, dose_per_kg) in [
        ["caffeine", 2],
        ["caffeine", 4],
        ["paraxanthine", 2],
        ["paraxanthine", 4],
    ]:

        print("substance: {}, dose: {}".format(substance, dose_per_kg))
        data = df[(df.substance == substance) & (df.dose == dose_per_kg)]

        # calculate pharmacokinetic information
        dose_per_kg = Q_(dose_per_kg, "mg/kg")
        bodyweight = Q_(70, "kg")
        dose = dose_per_kg * bodyweight  # [mg]
        t = data.time
        if substance == "caffeine":
            c = data.caf
        elif substance == "paraxanthine":
            c = data.px

        t = Q_(t.values, "hr")
        c = Q_(c.values, "mg/l")

        tcpk = TimecoursePK(
            time=t, concentration=c, substance=substance, dose=dose, ureg=ureg
        )

        results.append(tcpk)

    return results


def example_midazolam() -> List[TimecoursePK]:
    """Example for pharmacokinetics calculation.

    :return:
    """
    results = []
    df = pd.read_csv(TESTDATA_PATH / "pk" / "midazolam.tsv", sep="\t", na_values="NA")

    # ------------------------------------------
    # Pharmacokinetic parameter for acetaminophen
    # ------------------------------------------
    # get caffeine data
    dose = Q_(7.5, "mg")
    substance = "midazolam"

    # calculate pharmacokinetic information
    t = Q_(df.time.values, "min")
    c = Q_(df["Cve_mid"].values, "mmole/litre")
    tcpk = TimecoursePK(
        time=t, concentration=c, substance=substance, dose=dose, ureg=ureg
    )
    results.append(tcpk)

    return results


def example_Kim2011_Fig2() -> List[TimecoursePK]:
    """Example for pharmacokinetics calculation.

    :return:
    """
    results = []
    df = pd.read_csv(
        TESTDATA_PATH / "pk" / "Kim2011_Fig2.tsv", sep="\t", na_values="NA"
    )
    df = df[(df.interventions == "paracetamol1000mg")]

    # ------------------------------------------
    # Pharmacokinetic parameter for acetaminophen
    # ------------------------------------------
    # get caffeine data
    dose = Q_(100, "mg")
    substance = "acetaminophen"

    # calculate pharmacokinetic information
    t = Q_(df.time.values, "hr")
    c = Q_(df["mean"].values, "µg/ml")
    tcpk = TimecoursePK(
        time=t, concentration=c, substance=substance, dose=dose, ureg=ureg
    )
    results.append(tcpk)

    return results


def example_Divoll1982_Fig1() -> List[TimecoursePK]:
    """Example for pharmacokinetics calculation.

    :return:
    """
    results = []
    df = pd.read_csv(
        TESTDATA_PATH / "pk" / "Divoll1982_Fig1.tsv", sep="\t", na_values="NA"
    )

    # ------------------------------------------
    # Pharmacokinetic parameter for acetaminophen
    # ------------------------------------------
    # get caffeine data
    dose = Q_(100, "mg")
    substance = "acetaminophen"

    # calculate pharmacokinetic information
    t = Q_(df.time.values, "hr")
    c = Q_(df["apap"].values, "µg/ml")
    tcpk = TimecoursePK(
        time=t, concentration=c, substance=substance, dose=dose, ureg=ureg
    )
    results.append(tcpk)

    return results


def example_Lane2014_Fig1() -> List[TimecoursePK]:
    """Example for pharmacokinetics calculation.

    :return:
    """
    results = []
    df = pd.read_csv(
        TESTDATA_PATH / "pk" / "Lane2014_Fig1.tsv", sep="\t", na_values="NA"
    )

    # ------------------------------------------
    # Pharmacokinetic parameter for acetaminophen
    # ------------------------------------------
    # get caffeine data
    dose = Q_(0, "mg")
    substance = "caffeine"

    # calculate pharmacokinetic information
    t = Q_(df.time.values, "hr")
    c = Q_(df["mean"].values, "µg/ml")
    tcpk = TimecoursePK(
        time=t, concentration=c, substance=substance, dose=dose, ureg=ureg
    )
    results.append(tcpk)

    return results


def show_results(results: List[TimecoursePK]):
    """Show given results."""
    for tcpk in results:
        print(tcpk.info())
        f = tcpk.figure()
        plt.show()


if __name__ == "__main__":
    for f_example in [
        example0,
        example1,
        example1_NoDosing,
        example2,
        example_midazolam,
        example_Kim2011_Fig2,
        example_Divoll1982_Fig1,
        example_Lane2014_Fig1,
    ]:
        res = f_example()
        show_results(res)

    plt.show()



""" Defines frequently used filters for the PKData instances."""
from typing import List

from pkdb_analysis.data import PKData


TEST_STUDY_NAMES = ["Test1", "Test2", "Test3", "Test4"]
import collections
import warnings

import numpy as np
import pandas as pd


PlottingParameter = collections.namedtuple(
    "PlottingParameter", ["measurement_type", "units_rm"]
)


# add group_columns,
# add individual_columns,
# add intervention_columns,
# add output_columns,
# add timecourse_columns


def exclude_tests(data: PKData):
    return data.exclude_intervention(lambda d: d["study_name"].isin(TEST_STUDY_NAMES))


def combine(args):
    args = sorted(set([str(arg) for arg in args]))
    str_value = " || ".join(args)
    if str_value == "nan":
        return np.NAN
    try:
        return pd.to_numeric(str_value)

    except:
        return str_value


def pk_info(d, measurement_type, columns, suffix=None, concise=True, aggfunc=combine):
    if suffix is None:
        suffix_text = f"_{measurement_type}"
    else:
        suffix_text = suffix

    columns = [d.pk, *columns]
    df = (
        d[d["measurement_type"] == measurement_type][columns]
        .set_index(d.pk)
        .add_suffix(suffix_text)
        .reset_index()
    )
    if len(df) == 0:
        return df
    if concise:
        return df.pivot_table(index=d.pk, aggfunc=aggfunc, dropna=False).reset_index()
    return df


def f_unit(d, unit: str):
    """

    :param d: PKDataFrame or pandas.DataFrame
    :param unit:
    :return:
    """
    return d["unit"] == unit


def f_mt(d, measurement_type: str):
    """

    :param d: PKDataFrame or pandas.DataFrame
    :param measurement_type:
    :return:
    """
    return d["measurement_type"] == measurement_type


def f_substance(d, substance: str):
    """

    :param d: PKDataFrame or pandas.DataFrame
    :param substance:
    :return:
    """
    return d["substance"] == substance


def f_dosing(d, substance: str):
    """filtering index for PKData for dosing with substance

    :param d: PKDataFrame or pandas.DataFrame
    :param substance:
    :return:
    """
    return f_substance(d, substance) & f_mt(d, "dosing")


def f_dosing_in(d, substances: List[str]):
    """filtering index for PKData for dosing with substance

    :param d: PKDataFrame or pandas.DataFrame
    :param substance:
    :return:
    """
    d["substance"].isin(substances)
    return d["substance"].isin(substances) & f_mt(d, "dosing")


def f_mt_substance(d, measurement_type: str, substance: str):
    """filtering index for PKData for measurement_type with substance

    :param d: PKDataFrame or pandas.DataFrame
    :param measurement_type:
    :param substance:
    :return:
    """
    return f_mt(d, measurement_type) & f_substance(d, substance)


def f_mt_in_substance_in(d, measurement_types: str, substances: str):
    """filtering index for PKData for measurement_type with substance

    :param d: PKDataFrame or pandas.DataFrame
    :param measurement_type:
    :param substance:
    :return:
    """
    return d["measurement_type"].isin(measurement_types) & d["substance"].isin(
        substances
    )


def f_choice(d, choice):
    return d["choice"] == choice


def f_smoking(d):
    return f_mt(d, "smoking") & f_choice(d, "Y")


def f_n_smoking(d):
    return f_mt(d, "smoking") & f_choice(d, "N")


def f_oc(d):
    return f_mt(d, "oral contraceptives") & f_choice(d, "Y")


def f_n_oc(d):
    return f_mt(d, "oral contraceptives") & f_choice(d, "N")


def f_male(d):
    return f_mt(d, "sex") & f_choice(d, "M")


def f_female(d):
    return f_mt(d, "sex") & f_choice(d, "F")


def f_effective_n_oc(d):
    return (f_n_oc(d)) | (f_male(d))


def f_healthy(d):
    return f_mt(d, "healthy") & f_choice(d, "Y")


def f_n_healthy(d):
    return f_mt(d, "healthy") & f_choice(d, "N")



"""
Querying PK-DB
"""
import logging
import os
from copy import deepcopy
from pathlib import Path
from typing import List
from urllib import parse as urlparse

import numpy as np
import pandas as pd
import requests

from pkdb_analysis.data import PKData
from pkdb_analysis.envs import API_BASE, API_URL, PASSWORD, USER


logger = logging.getLogger(__name__)


def query_pkdb_data(
    h5_path: Path = None, username: str = None, study_names: List = None
) -> PKData:
    """Query the complete database.

    Filtering by study name is supported.

    :param filter_study_names: Iterable of study_names to filter for.
    :param username: filter studies by username
    """
    if h5_path.exists():
        logger.warning(f"Existing data file is overwritten: {h5_path}")

    if study_names is not None:
        study_filter = PKFilter()
        study_filter.add_to_all("study_name__in", "__".join(study_names))
        pkdata = PKDB.query(pkfilter=study_filter)
    else:
        pkdata = PKDB.query()

    if username is not None:
        # Filter studies by username
        pkdata.studies["username"] = pkdata.studies.creator.apply(pd.Series).username
        pkdata = pkdata.filter_study(
            f_idx=lambda pkdata: pkdata["username"] == username
        )

    if h5_path is not None:
        logger.info(f"Storing pkdb data: {h5_path}")
        pkdata.to_hdf5(h5_path)

    return pkdata


class PKFilter(object):
    """Filter objects for PKData"""

    KEYS = ["groups", "individuals", "interventions", "outputs", "data"]

    def __init__(self, normed=True):
        """Create new Filter instance.

        :param normed: [True, False, None] return [normed data, unnormalized data,
                        normed and unnormalized data]
        """
        self.groups = dict()
        self.individuals = dict()
        self.interventions = dict()
        self.outputs = dict()
        self.data = dict()
        self.studies = dict()

        self._set_normed(normed)

    def _set_normed(self, normed=True) -> None:
        """Set the normed attribute

        :param normed:
        :return: None
        """
        if normed not in [True, False, None]:
            raise ValueError

        if normed in [True, False]:
            if normed:
                normed_value = "true"
            else:
                normed_value = "false"
            for filter_key in ["interventions", "outputs"]:
                d = getattr(self, filter_key)
                d["normed"] = normed_value

    def __str__(self) -> str:
        return str(self.to_dict())

    def to_dict(self) -> dict:
        return {
            filter_key: deepcopy(getattr(self, filter_key))
            for filter_key in PKFilter.KEYS
        }

    def add_to_all(self, key, value) -> None:
        """Adds entry (key, value) to all KEY dictionaries

        :return: None
        """
        for filter_key in PKFilter.KEYS:
            # FIXME: THIS IS A REPLACE, NOT A ADD !!!!
            getattr(self, filter_key)[key] = value


class PKFilterFactory(object):
    """ Factory for simple creation of PKFilters. """

    @staticmethod
    def by_study_sid(study_sid: str) -> PKFilter:
        """Creates filter based on study_sid.
        Only data for the given study_sid is returned.
        """
        pkfilter = PKFilter()
        pkfilter.add_to_all("study_sid", study_sid)
        return pkfilter

    @staticmethod
    def by_study_name(study_name: str) -> PKFilter:
        """Creates filter based on study_name.
        Only data for the given study_name is returned.
        """
        pkfilter = PKFilter()
        pkfilter.add_to_all("study_name", study_name)
        return pkfilter


class PKDB(object):
    """ Querying PKData from PK-DB. """

    @classmethod
    def query(cls, pkfilter: PKFilter = PKFilter(), page_size: int = 2000) -> "PKData":
        """Create a PKData representation and gets the data for the provided filters.
        If no filters are given the complete data is retrieved.

        :param pkfilter: Filter object to select subset of data, if no Filter is provided the complete data is returned
        :param page_size: number of entries per query
        """
        pkfilter = pkfilter.to_dict()
        parameters = {"format": "json", "page_size": page_size}
        logger.info("*** Querying data ***")
        pkdata = PKData(
            studies=cls._get_subset(
                "studies", **{**parameters, **pkfilter.get("studies", {})}
            ),
            interventions=cls._get_subset(
                "interventions_analysis",
                **{**parameters, **pkfilter.get("interventions", {})},
            ),
            individuals=cls._get_subset(
                "individuals_analysis",
                **{**parameters, **pkfilter.get("individuals", {})},
            ),
            groups=cls._get_subset(
                "groups_analysis", **{**parameters, **pkfilter.get("groups", {})}
            ),
            outputs=cls._get_subset(
                "output_analysis", **{**parameters, **pkfilter.get("outputs", {})}
            ),
            data=cls._get_subset(
                "data_analysis", **{**parameters, **pkfilter.get("data", {})}
            ),
        )

        return cls._intervention_pk_update(pkdata)

    @classmethod
    def _get_subset(cls, name, **parameters):
        """

        :param name: name of the view
        :param parameters: query parameters
        :return:
        """
        if not name in [
            "individuals_analysis",  # individuals
            "groups_analysis",  # groups
            "interventions_analysis",  # interventions
            "output_analysis",  # outputs
            "data_analysis",  # data
            "studies",  # studies
        ]:
            raise ValueError(f"{name} not supported")

        url = API_URL + f"/{name}/"
        return cls._get_data(
            url, cls.get_authentication_headers(API_BASE, USER, PASSWORD), **parameters
        )

    @classmethod
    def get_authentication_headers(cls, api_base, username, password):
        """Get authentication header with token for given user.

        Returns admin authentication as default.
        """
        auth_dict = {"username": username, "password": password}
        for key, value in auth_dict.items():
            if value is None:
                return {}
        auth_token_url = urlparse.urljoin(api_base, "api-token-auth/")
        try:
            response = requests.post(auth_token_url, json=auth_dict)
        except requests.exceptions.ConnectionError as e:
            raise requests.exceptions.InvalidURL(
                f"Error Connecting (probably wrong url <{api_base}>): ", e
            )

        if response.status_code != 200:
            logger.error(
                f"Request headers could not be retrieved from: {auth_token_url}"
            )
            logger.warning(response.text)
            raise requests.exceptions.ConnectionError(response)

        token = response.json().get("token")
        return {"Authorization": f"token {token}"}

    @staticmethod
    def _get_data(url, headers, **parameters) -> pd.DataFrame:
        """Gets data from a paginated rest API."""
        url_params = "?" + urlparse.urlencode(parameters)
        actual_url = urlparse.urljoin(url, url_params)
        logger.info(actual_url)

        # FIXME: make first request fast
        response = requests.get(actual_url, headers=headers)
        try:
            response.raise_for_status()
            num_pages = response.json()["last_page"]

        except requests.exceptions.HTTPError as err:
            raise err

        data = []
        for page in range(1, num_pages + 1):
            url_current = actual_url + f"&page={page}"
            logger.info(url_current)

            response = requests.get(url_current, headers=headers)
            data += response.json()["data"]["data"]

        # convert to data frame
        df = pd.DataFrame(data)

        # convert columns to float columns
        float_columns = [
            "mean",
            "median",
            "value",
            "sd",
            "se",
            "cv",
            "min",
            "max",
            "time",
        ]

        for column in float_columns:
            if column in df.columns:
                df[column] = df[column].astype(float)

        # convert columns to int columns
        int_columns = [
            "intervention_pk",
            "group_pk",
            "individual_pk",
            "group_parent_pk",
            "raw_pk",
        ]
        for column in int_columns:
            if column in df.columns:
                df[column] = df[column].replace({np.nan: -1}).astype(int)

        return df

    @staticmethod
    def _map_intervention_pks(pkdata):
        interventions_output = pd.DataFrame()

        if not pkdata.outputs.empty:
            interventions_output = pkdata.outputs.df.pivot_table(
                values="intervention_pk",
                index="output_pk",
                aggfunc=lambda x: frozenset(x),
            )

        interventions = interventions_output.drop_duplicates(
            "intervention_pk"
        ).reset_index()
        interventions.index = interventions.index.set_names(["intervention_pk_updated"])
        return interventions["intervention_pk"].reset_index()

    @staticmethod
    def _update_interventions(pkdata, mapping_int_pks):
        mapping_int_pks = mapping_int_pks.copy()
        mapping_int_pks["intervention_pk"] = mapping_int_pks.intervention_pk.apply(
            lambda x: list(x)
        )
        mapping_int_pks = (
            mapping_int_pks.intervention_pk.apply(pd.Series)
            .stack()
            .reset_index(level=-1, drop=True)
            .astype(int)
            .reset_index()
        )
        mapping_int_pks = mapping_int_pks.rename(
            columns={"index": "intervention_pk_updated", 0: "intervention_pk"}
        )
        return (
            pd.merge(mapping_int_pks, pkdata.interventions, on="intervention_pk")
            .drop(columns=["intervention_pk"])
            .rename(columns={"intervention_pk_updated": "intervention_pk"})
        )

    @staticmethod
    def _update_outputs(pkdata, mapping_int_pks):
        mapping_int_pks = mapping_int_pks.copy()

        interventions_output = pkdata.outputs.df.pivot_table(
            values="intervention_pk", index="output_pk", aggfunc=lambda x: frozenset(x)
        )
        mapping_int_pks = pd.merge(
            interventions_output.reset_index(),
            mapping_int_pks,
            on="intervention_pk",
            how="left",
        )[["output_pk", "intervention_pk_updated"]]

        return (
            pd.merge(
                mapping_int_pks,
                pkdata.outputs.df.drop_duplicates(subset="output_pk"),
                how="left",
            )
            .drop(columns=["intervention_pk"])
            .rename(columns={"intervention_pk_updated": "intervention_pk"})
        )

    @classmethod
    def _intervention_pk_update(cls, pkdata):
        if not pkdata.outputs.empty:
            mapping_int_pks = cls._map_intervention_pks(pkdata)

            data_dict = pkdata.as_dict()
            data_dict["interventions"] = cls._update_interventions(
                pkdata, mapping_int_pks
            )
            if not pkdata.outputs.empty:
                data_dict["outputs"] = cls._update_outputs(pkdata, mapping_int_pks)
            return PKData(**data_dict)
        else:
            return pkdata



import numpy as np
import pandas as pd

from pkdb_analysis.analysis import figure_category
from pkdb_analysis.filter import f_healthy, f_n_healthy, pk_info
from pkdb_analysis.inference.body_weight import infer_weight


INTERVENTION_FIELDS = ["substance", "value", "unit", "route", "form", "application"]
NUMERIC_FIELDS_NO_VALUE = ["mean", "min", "max", "median", "count", "sd", "se", "unit"]
NUMERIC_FIELDS = ["value"] + NUMERIC_FIELDS_NO_VALUE


def one_intervention(d):
    if len(d) == 1:
        return d


def validate_1len(d):
    assert 1 == len(d[d]), d
    return d


def markers(d):
    if d["calculated"]:
        return "s"
    elif d["inferred"]:
        return "v"
    else:
        return "o"


def data_type(d):
    if d["calculated"]:
        return "from timecourse"
    elif d["inferred"]:
        return "from bodyweight"
    else:
        return "publication"


class MetaAnalysis(object):
    def __init__(self, pkdata, intervention_substances, url):
        self.pkdata = pkdata
        self.results = None
        self.group_pk = pkdata.groups.pk
        self.individual_pk = pkdata.individuals.pk
        self.intervention_pk = pkdata.interventions.pk
        self.intervention_substances = intervention_substances
        self.url = url

    def _create_extra_table(self, table_name, substances):

        table = getattr(self.pkdata, table_name)
        _table = pd.DataFrame()
        for table_pk, df in table.df.groupby(table.pk):
            subset = df[df["substance"].isin(substances)]
            subset["number"] = len(df)
            if len(subset) == 1:
                subset = subset.iloc[0]
                subset["extra"] = df
                _table = _table.append(subset)
        return _table

    def _add(self, df, measurement_type):
        age_data = df.extra[df.extra["measurement_type"] == measurement_type]
        if len(age_data) == 1:
            return tuple(age_data.iloc[0][NUMERIC_FIELDS].values)
        else:
            return tuple([np.nan for _ in NUMERIC_FIELDS])

    @property
    def healthy_data(self):
        return self.pkdata.filter_subject(f_healthy).exclude_subject(f_n_healthy)

    def create_subject_table(
        self,
        subject,
        numeric_fields=("weight", "age"),
        catgorial_fields=("sex",),
        add_healthy=True,
    ):

        subject_core = getattr(self.pkdata, f"{subject}_core")
        subject_df = getattr(self.pkdata, subject)

        if add_healthy:
            healthy_subjects_pks = getattr(self.healthy_data, subject).pks
            subject_core["healthy"] = subject_core[subject_df.pk].isin(
                healthy_subjects_pks
            )

        subject_numeric_extra = [
            pk_info(subject_df, numeric_field, NUMERIC_FIELDS)
            for numeric_field in numeric_fields
        ]
        subject_categorials_extra = []
        for categorial_field in catgorial_fields:
            detail_info = pk_info(subject_df, categorial_field, ["choice"]).rename(
                columns={f"choice_{categorial_field}": categorial_field}
            )
            subject_categorials_extra.append(detail_info)

        for individuals_info in [*subject_categorials_extra, *subject_numeric_extra]:
            subject_core = pd.merge(
                subject_core, individuals_info, on=subject_df.pk, how="left"
            )

        for categorial_field in catgorial_fields:
            subject_core[categorial_field] = subject_core[categorial_field].fillna(
                "unknown"
            )

        pk = subject_df.pk
        subject_core["extra"] = getattr(subject_core, pk).apply(
            lambda x: subject_df[subject_df[pk] == x]
        )
        for measurement_type in ["age", "weight"]:
            subject_extra = subject_core.apply(
                self._add, args=(measurement_type,), axis=1, result_type="expand"
            )
            if subject_extra.empty:
                subject_core[
                    [f"{k}_{measurement_type}" for k in NUMERIC_FIELDS]
                ] = tuple([np.nan for _ in NUMERIC_FIELDS])
            else:
                subject_core[
                    [f"{k}_{measurement_type}" for k in NUMERIC_FIELDS]
                ] = subject_extra

        return subject_core

    def add_extra_info(self, replacements):
        self.results["unit_category"] = self.results[
            ["per_bw", "intervention_per_bw"]
        ].apply(figure_category, axis=1)
        self.results["y"] = self.results[["mean", "median", "value"]].max(axis=1)
        self.results["y_min"] = self.results["y"] - self.results["sd"]
        self.results["y_max"] = self.results["y"] + self.results["sd"]

        self.results["weight"] = self.results[
            ["mean_weight", "median_weight", "value_weight"]
        ].max(axis=1)
        self.results["min_sd_weight"] = (
            self.results["weight"] - self.results["sd_weight"]
        )
        self.results["max_sd_weight"] = (
            self.results["weight"] + self.results["sd_weight"]
        )

        self.results["age"] = self.results[["mean_age", "median_age", "value_age"]].max(
            axis=1
        )
        self.results["min_sd_age"] = self.results["age"] - self.results["sd_age"]
        self.results["max_sd_age"] = self.results["age"] + self.results["sd_age"]

        self.results["subject_count"] = self.results["group_count"].fillna(1)

        self.results["url"] = self.results["study_sid"].apply(
            lambda x: f"{self.url}/studies/{x}"
        )
        # self.results = self.results.replace({"NR", "not reported"}, regex=True)

        for column, replace_dict in replacements.items():
            self.results[column] = self.results[column].replace(replace_dict)

    def create_results_base(self):
        results = self.pkdata.outputs.copy()
        results["per_bw"] = results.unit.str.endswith("/ kilogram")
        results["inferred"] = False
        self.results = results

    def add_intervention_info(self):
        intervention_table = self._create_extra_table(
            "interventions", self.intervention_substances
        )
        intervention_table["per_bw"] = intervention_table.unit.str.endswith(
            "/ kilogram"
        )
        intervention_table = intervention_table.rename(
            columns={"intervention_pk": "pk"}
        )
        self.results = pd.merge(
            self.results,
            intervention_table.add_prefix("intervention_"),
            on=self.intervention_pk,
            how="inner",
        )

    def individual_results(self):
        individual = "individuals"
        individual_table = self.create_subject_table(individual)
        individual_results = self.results[self.results.group_pk == -1]
        unique_individual_columns = set(individual_table.columns).difference(
            individual_results.columns
        )

        unique_individual_columns = [*unique_individual_columns, self.individual_pk]

        return pd.merge(
            individual_results,
            individual_table[unique_individual_columns],
            on=self.individual_pk,
            how="left",
        )

    def group_results(self):
        group = "groups"
        group_table = self.create_subject_table(group)
        group_results = self.results[self.results.group_pk != -1]
        unique_group_columns = set(group_table.columns).difference(
            group_results.columns
        )
        unique_group_columns = [*unique_group_columns, self.group_pk]
        return pd.merge(
            group_results,
            group_table[unique_group_columns],
            on=self.group_pk,
            how="left",
        )

    def add_subject_info(self):

        self.results = self.individual_results().df.append(self.group_results())

    def infer_from_body_weight(self, by_intervention=True, by_output=True):
        results_inferred = infer_weight(self.results, by_intervention, by_output)
        results_inferred["marker"] = results_inferred.apply(markers, axis=1)
        results_inferred["data_type"] = results_inferred.apply(data_type, axis=1)
        self.results = results_inferred

    def create_results(self):
        self.create_results_base()
        self.add_intervention_info()
        self.add_subject_info()



import pandas as pd
import pytest
import requests

from pkdb_analysis import PKFilterFactory
from pkdb_analysis.data import PKData
from pkdb_analysis.envs import API_BASE, PASSWORD, USER
from pkdb_analysis.query import PKDB


def _check_data(data):
    assert data

    for key in ["groups", "individuals", "interventions", "outputs", "timecourses"]:
        df = getattr(data, key)

        assert df is not None
        assert isinstance(df, pd.DataFrame)
        assert len(df) > 0
        assert "study_sid" in df


def _check_data_empty(data):
    assert data

    for key in ["groups", "individuals", "interventions", "outputs", "timecourses"]:
        df = getattr(data, key)

        assert df is not None
        assert isinstance(df, pd.DataFrame)
        assert len(df) == 0
        assert "study_sid" not in df


def is_admin_and_connection():
    if USER == "admin":
        try:
            PKDB.get_authentication_headers(API_BASE, USER, PASSWORD)
            return False
        except requests.exceptions.InvalidURL:
            return True
    return True


@pytest.mark.skipif(
    is_admin_and_connection(), reason="Database not running or not admin user."
)
def test_data_by_study_name():
    # check existing study
    pkfilter = PKFilterFactory.by_study_name("Test1")
    data = PKDB.query(pkfilter=pkfilter)
    _check_data(data)

    for key in PKData.KEYS:
        df = getattr(data, key)
        study_names = df.study_name.unique()
        assert len(study_names) == 1
        assert "Test1" in study_names


@pytest.mark.skipif(
    is_admin_and_connection(), reason="Database not running or not admin user."
)
def test_data_by_study_name_empty():
    # check non-existing study
    pkfilter = PKFilterFactory.by_study_name("xyzfasdfs")
    data = PKDB.query(pkfilter=pkfilter)
    _check_data_empty(data)


@pytest.mark.skipif(
    is_admin_and_connection(), reason="Database not running or not admin user."
)
def test_data_by_study_sid():
    # check existing study
    pkfilter = PKFilterFactory.by_study_sid("Test1")
    data = PKDB.query(pkfilter=pkfilter)

    _check_data(data)

    for key in PKData.KEYS:
        df = getattr(data, key)
        study_sids = df.study_sid.unique()
        assert len(study_sids) == 1
        assert "Test1" in study_sids


@pytest.mark.skipif(
    is_admin_and_connection(), reason="Database not running or not admin user."
)
def test_data_by_study_sid_empty():
    # check non-existing study
    pkfilter = PKFilterFactory.by_study_sid("xyzfasdfs")
    data = PKDB.query(pkfilter=pkfilter)
    _check_data_empty(data)


@pytest.mark.skipif(
    is_admin_and_connection(), reason="Database not running or not admin user."
)
def test_data_hdf5(tmp_path):
    """ Test HDF io"""
    data = PKDB.query()

    h5_path = tmp_path / "test.h5"
    data.to_hdf5(h5_path)
    data2 = PKData.from_hdf5(h5_path)
    _check_data(data2)

    for key in PKData.KEYS:
        assert len(getattr(data, key)) == len(getattr(data2, key))


@pytest.mark.skipif(
    is_admin_and_connection(), reason="Database not running or not admin user."
)
def test_data_counts():
    pkfilter = PKFilterFactory.by_study_name("Test1")
    data = PKDB.query(pkfilter=pkfilter)
    assert data.groups_count == 1
    assert data.individuals_count == 6
    assert data.interventions_count == 3
    assert data.outputs_count == 105
    assert data.timecourses_count == 2


@pytest.mark.skipif(
    is_admin_and_connection(), reason="Database not running or not admin user."
)
def test_data_mi():
    pkfilter = PKFilterFactory.by_study_name("Test1")
    data = PKDB.query(pkfilter=pkfilter)
    for field in PKData.KEYS:
        print(field)
        df = getattr(data, f"{field}_mi")
        assert df is not None
        assert not df.empty
        # number of entries should not change
        assert len(df) == len(getattr(data, f"{field}"))


@pytest.mark.skipif(
    is_admin_and_connection(), reason="Database not running or not admin user."
)
def test_data_test1():
    # check non-existing study
    pkfilter = PKFilterFactory.by_study_name("Test1")
    data = PKDB.query(pkfilter=pkfilter)
    _check_data(data)

    assert len(data.groups) == 6



"""
Functions for working with PKDB data.

* FIXME: specify which methods modify or copy data frames
"""
import logging
import warnings
from abc import ABC
from collections import OrderedDict
from copy import copy
from pathlib import Path
from typing import Callable, List

import numpy as np
import pandas as pd
from IPython.display import display
from pandas.errors import PerformanceWarning


warnings.simplefilter(action="ignore", category=PerformanceWarning)
logger = logging.getLogger(__name__)


class PKDataFrame(pd.DataFrame, ABC):
    """
    Extended DataFrame which support customized filter operations.
    Used to encode groups, individuals, interventions, outputs, data on PKData.
    """

    @property
    def _constructor(self):
        return PKDataFrame._internal_ctor

    _metadata = ["pk"]

    @classmethod
    def _internal_ctor(cls, *args, **kwargs):
        kwargs["pk"] = None
        return cls(*args, **kwargs)

    def __init__(self, data, pk=None, index=None, columns=None, dtype=None, copy=True):
        """

        :param data:
        :param pk:
        :param index:
        :param columns:
        :param dtype:
        :param copy:
        """
        if not isinstance(data, pd.core.internals.BlockManager) and not pk:
            raise ValueError("arg pk required")

        super(PKDataFrame, self).__init__(
            data=data, index=index, columns=columns, dtype=dtype, copy=copy
        )
        self.pk = pk

    @staticmethod
    def _validate_not_in_columns(columns):
        if columns:
            assert "pk" in columns

    def pk_filter(self, f_idx: Callable, **kwargs) -> "PKDataFrame":
        """

        :param f_idx: function, index, list
        :return:
        """

        def ff_idx(d):
            return f_idx(d, **kwargs)

        if isinstance(f_idx, list):

            pk_df = self
            for f_idx_single in f_idx:
                pk_df = pk_df.pk_filter(f_idx_single, **kwargs)
            return pk_df

        df_pks = self[ff_idx][self.pk].unique()
        df_filtered = self[self[self.pk].isin(df_pks)]
        return PKDataFrame(df_filtered, pk=self.pk)

    def pk_exclude(self, f_idx: Callable, **kwargs) -> "PKDataFrame":
        def ff_idx(d):
            return f_idx(d, **kwargs)

        if isinstance(f_idx, list):
            pk_df = self
            for f_idx_single in f_idx:
                pk_df = pk_df.pk_exclude(f_idx_single, **kwargs)
            return pk_df

        df_pks = self[ff_idx][self.pk].unique()
        df_excluded = self[~self[self.pk].isin(df_pks)]
        return PKDataFrame(df_excluded, pk=self.pk)

    @property
    def pk_column(self):
        return self[self.pk]

    @property
    def pks(self) -> set:
        """ Set of pks."""
        if self.pk in self.df.columns:
            return set(self[self.pk].unique())
        else:
            return set()

    @property
    def pk_len(self) -> int:
        return len(self.pks)

    @property
    def df(self) -> pd.DataFrame:
        """ Returns a copied DataFrame."""
        df = self.copy()
        del df.pk
        return pd.DataFrame(self)

    @property
    def study_sids(self) -> set:
        """ Set of study_sids."""
        study_sids = set([])
        if "study_sid" in self.df.columns:
            study_sids = set(self.study_sid.unique())
        return study_sids

    def _emptify(self) -> "PKDataFrame":
        empty_df = pd.DataFrame([], columns=self.columns)
        return PKDataFrame(empty_df, pk=self.pk)

    def __str__(self):
        return self.df.__str__()

    def __repr__(self):
        return self.df.__repr__()

    def _repr_html_(self):
        return display(self.df)


class PKData(object):
    PK_COLUMNS = {
        "studies": "study_pk",
        "groups": "group_pk",
        "individuals": "individual_pk",
        "interventions": "intervention_pk",
        "outputs": "output_pk",
        "data": "subset_pk",
    }
    """ Consistent set of data from PK-DB.set -a && source .env.local

    Information is stored as DataFrames.

    Handles:

    - groups
    - individuals
    - interventions
    - outputs
    - data
    """
    KEYS = ["studies", "groups", "individuals", "interventions", "outputs", "data"]

    def __init__(
        self,
        interventions: pd.DataFrame = None,
        groups: pd.DataFrame = None,
        individuals: pd.DataFrame = None,
        outputs: pd.DataFrame = None,
        data: pd.DataFrame = None,
        studies: pd.DataFrame = None,
    ):
        """Creates PKDB data object from given DataFrames.

        :param interventions:
        :param individuals:
        :param groups:
        :param outputs:
        :param data:
        """
        self.groups = PKDataFrame(groups, pk="group_pk")
        self.individuals = PKDataFrame(individuals, pk="individual_pk")
        self.interventions = PKDataFrame(interventions, pk="intervention_pk")
        self.outputs = PKDataFrame(outputs, pk="output_pk")
        self.data = PKDataFrame(data, pk="subset_pk")
        self.studies = PKDataFrame(studies, pk="sid")

        if not self.individuals.empty:
            self.individuals.substance = self.individuals.substance.astype(str)
        if not self.groups.empty:
            self.groups.substance = self.groups.substance.astype(str)

    def __dict___(self):
        return {df_key: getattr(self, df_key).df for df_key in PKData.KEYS}

    def as_dict(self):
        return self.__dict___()

    def copy(self):
        return PKData(**self.as_dict())

    def __str__(self):
        """Overview of content.

        :return:
        """
        lines = [
            "-" * 30,
            f"{self.__class__.__name__} ({id(self)})",
            "-" * 30,
        ]

        for key in self.KEYS:
            if key != "data":
                df = getattr(self, key)
                nrows = len(df)
                count = df.pk_len
                lines.append(f"{key:<15} {count:>5}  ({nrows:>5})")
            else:
                lines.append(
                    f"{key:<12} tc:{ self.timecourses_count:>5}  sc:{self.scatter_count:>3}"
                )

        lines.append("-" * 30)
        return "\n".join(lines)

    def __or__(self, other: "PKData") -> "PKData":
        """combines two PKData instances
        :param other: other PkData instance
        :return: PKData
        """

        resulting_kwargs = dict()

        for df_key in self.KEYS:
            df = getattr(self, df_key)
            other_df = getattr(other, df_key)
            resulting_df = df.append(other_df)
            resulting_df = resulting_df.loc[
                ~resulting_df.index.duplicated(keep="first")
            ]

            resulting_kwargs[df_key] = resulting_df

        return PKData(**resulting_kwargs)

    def __and__(self, other: "PKData") -> "PKData":
        """combines instances were instances have to

        param other: other PKData instance
        :return: PKData
        """

        resulting_kwargs = dict()
        for df_key in PKData.KEYS:
            df = getattr(self, df_key)
            other_df = getattr(other, df_key)

            pk = PKData.PK_COLUMNS
            pks = set(df[pk])
            other_pks = set(other_df[pk])

            intersection_pks = pks.intersection(other_pks)
            df = df[df[pk].isin(intersection_pks)]
            other_df = other_df[other_df[pk].isin(intersection_pks)]

            resulting_df = df.append(other_df)
            resulting_df = resulting_df.loc[
                ~resulting_df.index.duplicated(keep="first")
            ]
            resulting_kwargs[df_key] = resulting_df

        return PKData(**resulting_kwargs)

    @staticmethod
    def from_hdf5(path: Path) -> "PKData":
        """Load data from HDF5 serialization.

        :param path: path to HDF5.
        :type path: str
        :return: PKData loaded from HDF5.
        :rtype: PKData
        """
        store = pd.HDFStore(path)
        data_dict = {}
        for key in store.keys():
            # ugly bugfix due to hdf5 key mutation (key -> /key on storage)
            data_dict[key[1:]] = store[key]
        store.close()

        return PKData(**data_dict)

    def to_hdf5(self, path: Path) -> None:
        """Saves data HDF5

        :param path: path to HDF5.
        :type path: str
        """
        store = pd.HDFStore(path)
        for key in [
            "studies",
            "interventions",
            "individuals",
            "groups",
            "outputs",
            "data",
        ]:
            df = getattr(self, key).df
            store[key] = df
        store.close()

    @property
    def study_sids(self) -> set:
        """Set of study sids contained in this PKData instance.

        :return: Study sids contained in this PKData instance.
        :rtype: set
        """
        study_sids = set()
        for df_key in PKData.KEYS:
            pk_df = getattr(self, df_key)
            study_sids = study_sids.union(pk_df.study_sids)

        return study_sids

    @property
    def groups_count(self) -> int:
        """Number of groups contained in this PKData instance.

        :return: Number of groups contained in this PKData instance.
        :rtype: int
        """
        return self.groups.pk_len

    @property
    def individuals_count(self) -> int:
        """Number of individuals contained in this PKData instance.

        :return: Number of individuals contained in this PKData instance.
        :rtype: int
        """
        return self.individuals.pk_len

    @property
    def interventions_count(self) -> int:
        """Number of interventions contained in this PKData instance.

        :return: Number of interventions contained in this PKData instance.
        :rtype: int
        """
        return self.interventions.pk_len

    @property
    def outputs_count(self) -> int:
        """Number of outputs contained in this PKData instance.

        :return: Number of outputs contained in this PKData instance.
        :rtype: int
        """
        return self.outputs.pk_len

    @property
    def timecourses_count(self) -> int:
        """Number of timecourses contained in this PKData instance.

        :return: Number of timecourses contained in this PKData instance.
        :rtype: int
        """
        return self.data[self.data.data_type == "timecourse"].pk_len

    @property
    def scatter_count(self) -> int:
        """Number of timecourses contained in this PKData instance.

        :return: Number of timecourses contained in this PKData instance.
        :rtype: int
        """
        return self.data[self.data.data_type == "scatter"].pk_len

    @property
    def timecourses_extended(self) -> pd.DataFrame:
        """ extends the timecourse df with the core information from interventions, individuals and groups"""

        timecourses = self.timecourses.df.merge(
            self.interventions_core,
            how="left",
            on="intervention_pk",
            suffixes=("", "interventions"),
        )
        timecourses = timecourses.merge(
            self.individuals_core,
            how="left",
            on="individual_pk",
            suffixes=("", "individuals"),
        )
        timecourses = timecourses.merge(
            self.groups_core, how="left", on="group_pk", suffixes=("", "groups")
        )
        return timecourses

    def _df_mi(self, field: str, index_fields: List[str]) -> pd.DataFrame:
        """Create multi-index DataFrame

        :param field:  #fixme rename to df_key?
        :param index_fields:
        :return: Multi-indexed Dataframe
        :rtype: pd.DataFrame
        """
        df = getattr(self, field)
        if df.empty:
            return pd.DataFrame()  # new empty DataFrame
        # create multi-index DataFrame
        df_mi = df.sort_values(index_fields, ascending=True, inplace=False)
        df_mi.set_index(index_fields, inplace=True)
        return df_mi

    def _df_core(self, field: str, core_fields: List[str]) -> PKDataFrame:
        """Core group information with unique pk per row"""

        pk_df = getattr(self, field)
        return pk_df.pivot_table(
            index=pk_df.pk, values=core_fields, aggfunc=lambda x: x.iloc[0]
        ).reset_index()

    @property
    def groups_mi(self) -> pd.DataFrame:
        """Multi-index DataFrame of groups contained in this PKData instance.

        :return: Multi-indexed DataFrame of groups contained in this PKData instance.
        :rtype: pd.DataFrame
        """
        return self._df_mi("groups", ["group_pk", "characteristica_pk"])

    @property
    def groups_core(self) -> PKDataFrame:
        """Core group information with unique pk per row

        :return: PKDataFrame of groups contained in this PKData instance .
        :rtype: PKDataFrame
        """
        return self._df_core(
            "groups", core_fields=["study_name", "group_name", "group_count"]
        )

    @property
    def individuals_mi(self) -> pd.DataFrame:
        """Multi-index DataFrame of individuals contained in this PKData instance.

        :return: Multi-indexed DataFrame of individuals contained in this PKData instance.
        :rtype: pd.DataFrame
        """
        return self._df_mi(
            "individuals", ["individual_pk", "individual_name", "characteristica_pk"]
        )

    @property
    def individuals_core(self) -> PKDataFrame:
        """Core individual information with unique pk per row

        :return: PKDataFrame of individuals contained in this PKData instance .
        :rtype: PKDataFrame
        """
        return self._df_core(
            "individuals", core_fields=["study_name", "individual_name"]
        )

    @property
    def interventions_mi(self) -> pd.DataFrame:
        """Multi-index DataFrame of interventions contained in this PKData instance.

        :return: Multi-indexed DataFrame of interventions contained in this PKData instance.
        :rtype: pd.DataFrame
        """
        return self._df_mi("interventions", ["intervention_pk"])

    @property
    def interventions_core(self) -> PKDataFrame:
        """Core group information with unique pk per row

        :return: PKDataFrame of groups contained in this PKData instance .
        :rtype: PKDataFrame
        """
        return self._df_core("interventions", core_fields=["study_name", "name"])

    @property
    def outputs_mi(self) -> pd.DataFrame:
        """Multi-index DataFrame of outputs contained in this PKData instance.

        :return: Multi-indexed DataFrame of outputs contained in this PKData instance.
        :rtype: pd.DataFrame
        """
        return self._df_mi(
            "outputs", ["output_pk", "intervention_pk", "group_pk", "individual_pk"]
        )

    # --- filter and exclude ---
    def _pk_filter(
        self, df_key: str, f_idx, concise: bool, *args, **kwargs
    ) -> "PKData":
        """Helper class for filtering of PKData instances.
        :param df_key: DataFrame on which the filter (f_idx) shall be applied.
        :type df_key: str
        :param concise:
        :return:
        """
        dict_pkdata = self.as_dict()
        dict_pkdata[df_key] = getattr(self, df_key).pk_filter(f_idx, **kwargs)

        pkdata = PKData(**dict_pkdata)
        if concise:
            pkdata._concise()
        return pkdata

    def _pk_exclude(self, df_k, f_idx, concise, **kwargs) -> "PKData":

        dict_pkdata = self.as_dict()
        dict_pkdata[df_k] = getattr(self, df_k).pk_exclude(f_idx, **kwargs)

        pkdata = PKData(**dict_pkdata)
        if concise:
            pkdata._concise()
        return pkdata

    def _emptify(self, df_key, concise=True) -> "PKData":
        self._validate_df_key(df_key)

        dict_pkdata = self.as_dict()
        dict_pkdata[df_key] = getattr(self, df_key)._emptify()

        pkdata = PKData(**dict_pkdata)
        if concise:
            pkdata._concise()
        return pkdata

    @staticmethod
    def _validate_df_key(df_key):
        if df_key not in PKData.KEYS:
            raise ValueError(
                f"Unsupported key '{df_key}', key must be in '{PKData.KEYS}'"
            )

    def filter_study(self, f_idx, concise=True, **kwargs) -> "PKData":
        """ Filter studies by filter function. """
        return self._pk_filter("studies", f_idx, concise, **kwargs)

    def filter_intervention(self, f_idx, concise=True, *args, **kwargs) -> "PKData":
        """Filter interventions.

        :param f_idx: Is a filter by index of the DataFrame selected by the df_key. A similar notation as the filtering
            of pd.DataFrames can be used. This mostly are (lambda) functions. Further a list of (lambda) function are allowed
            as input. The list of functions are executed successively, which is identical to an intersection of all filters
            applied separately.

            Pitfalls
                - Don't use the invert operator `~` but use the exclude_*() functions.
                - #todo: add no invert operator to Validation rule
        :type f_idx: function, list


        :param concise: concises the returned PKData instance.

        :return: Filter PKData instance
        :rtype: PKData
        """

        return self._pk_filter("interventions", f_idx, concise, **kwargs)

    def filter_group(self, f_idx, concise=True, **kwargs) -> "PKData":
        """ Filter groups. """
        return self._pk_filter("groups", f_idx, concise, **kwargs)

    def filter_individual(self, f_idx, concise=True, **kwargs) -> "PKData":

        """ Filter individuals. """
        return self._pk_filter("individuals", f_idx, concise, **kwargs)

    def filter_subject(self, f_idx, concise=True, **kwargs) -> "PKData":
        """ Filter group or individual. """
        pkdata = self.filter_group(f_idx, concise=False, **kwargs)
        pkdata = pkdata.filter_individual(f_idx, concise=False, **kwargs)
        if concise:
            pkdata._concise()
        return pkdata

    def filter_output(self, f_idx, concise=True, **kwargs) -> "PKData":
        """ Filter outputs. """
        return self._pk_filter("outputs", f_idx, concise, **kwargs)

    def filter_data(self, f_idx, concise=True, **kwargs) -> "PKData":
        """ Filter data. """
        return self._pk_filter("data", f_idx, concise, **kwargs)

    def exclude_study(self, f_idx, concise=True, **kwargs) -> "PKData":
        """ Filter groups. """
        return self._pk_exclude("studies", f_idx, concise, **kwargs)

    def exclude_intervention(self, f_idx, concise=True, **kwargs) -> "PKData":
        return self._pk_exclude("interventions", f_idx, concise, **kwargs)

    def exclude_group(self, f_idx, concise=True, **kwargs) -> "PKData":
        return self._pk_exclude("groups", f_idx, concise, **kwargs)

    def exclude_individual(self, f_idx, concise=True, **kwargs):
        return self._pk_exclude("individuals", f_idx, concise, **kwargs)

    def exclude_subject(self, f_idx, concise=True, **kwargs) -> "PKData":
        pkdata = self.exclude_group(f_idx, concise=False, **kwargs)
        pkdata = pkdata.exclude_individual(f_idx, concise=False, **kwargs)
        if concise:
            pkdata._concise()
        return pkdata

    def exclude_output(self, f_idx, concise=True, **kwargs) -> "PKData":
        return self._pk_exclude("outputs", f_idx, concise, **kwargs)

    def exclude_data(self, f_idx, concise=True, **kwargs):

        return self._pk_exclude("data", f_idx, concise, **kwargs)

    def delete_groups(self, concise=True) -> "PKData":
        """
        Deletes outputs.
        :return:
        """
        return self._emptify("groups", concise=concise)

    def delete_individuals(self, concise=True) -> "PKData":
        """
        Deletes outputs.
        :return:
        """
        return self._emptify("individuals", concise=concise)

    def delete_outputs(self, concise=True) -> "PKData":
        """
        Deletes outputs.
        :return:
        """
        return self._emptify("outputs", concise=concise)

    def delete_data(self, concise=True) -> "PKData":
        """Deletes data."""
        return self._emptify("data", concise=concise)

    def _concise(self) -> None:
        """Reduces the current PKData to a consistent subset.
        Modifies the DataFrame in place.
        :return:
        """
        previous_len = np.inf
        logger.warning("Concise DataFrames")
        while previous_len > self._len_total:
            previous_len = copy(self._len_total)

            # concise based on studies
            outputs_study_sids = set(self.outputs.study_sid)
            data_study_sids = set(self.data.study_sid)
            interventions_study_sids = set(self.interventions.study_sid)
            groups_study_sids = set(self.groups.study_sid)
            individuals_study_sids = set(self.individuals.study_sid)

            study_sid_sets = [
                outputs_study_sids,
                data_study_sids,
                interventions_study_sids,
                groups_study_sids,
                individuals_study_sids,
            ]

            current_study_sid_sets = set().union(*study_sid_sets)
            self.studies = self.studies[self.studies.sid.isin(current_study_sid_sets)]
            for df_key in ["interventions", "groups", "individuals", "data", "outputs"]:
                df = getattr(self, df_key)
                setattr(self, df_key, df[df["study_sid"].isin(self.studies.pks)])

            # concise based on interventions
            outputs_intervention_pks = set(self.outputs.intervention_pk)
            current_intervention_pks = self.interventions.pks.intersection(
                outputs_intervention_pks
            )
            current_intervention_pks.add(-1)

            self.interventions = self.interventions[
                self.interventions.intervention_pk.isin(current_intervention_pks)
            ]
            self.outputs = self.outputs[
                self.outputs.intervention_pk.isin(current_intervention_pks)
            ]

            # concise based on individuals
            outputs_individual_pks = set(self.outputs.individual_pk)
            current_individual_pks = self.individuals.pks.intersection(
                outputs_individual_pks
            )
            current_individual_pks.add(-1)

            for df_key in ["individuals", "outputs"]:
                df = getattr(self, df_key)
                setattr(
                    self, df_key, df[df["individual_pk"].isin(current_individual_pks)]
                )

            # concise based on groups
            outputs_group_pks = set(self.outputs.group_pk)

            current_group_pks = self.groups.pks.intersection(outputs_group_pks)
            current_group_pks.add(-1)

            for df_key in ["groups", "outputs"]:
                df = getattr(self, df_key)
                setattr(self, df_key, df[df.group_pk.isin(current_group_pks)])

    @property
    def _len_total(self):
        return sum([len(getattr(self, df_key)) for df_key in PKData.KEYS])

    def get_choices(self):
        """This is experimental.
        returns choices

        :return:
        """
        all_choices = OrderedDict()
        for df_key in self.KEYS:
            df = getattr(self, df_key)
            choices = OrderedDict()
            for key in df.columns:
                if df[key].dtype in ["bool", "object"]:
                    # remove None so sorting is working
                    values = [c for c in set(df[key]) if c is not None]
                    choices[key] = sorted(values)

            all_choices[df_key] = choices
        return all_choices

    def print_choices(self, key=None, field=None):
        """Prints the choices

        :param key: key of dataframe
        :param field: header field
        :return:
        """
        if key is None:
            df_keys = PKData.KEYS
        else:
            PKData._validate_df_key(key)
            df_keys = [key]

        all_choices = self.choices
        for df_key in df_keys:
            choices = all_choices[df_key]
            if field is not None:
                if field not in choices.keys():
                    raise ValueError(
                        f"Unsupported field '{field}', field must be in '{choices.keys()}'"
                    )
                fields = [field]
            else:
                fields = choices.keys()

            for field in fields:
                print(f"*** {field} ***")
                print(choices[field])



from pathlib import Path

from pkdb_analysis.query import query_pkdb_data
from pkdb_analysis.test import TEST_HDF5


def create_h5_test_data(h5_path: Path = TEST_HDF5) -> None:
    """Loads test studies from database.

    Ensure test studies are uploaded to database before running this script.
        ./pkdb_data/management/upload_studies.py tests

    :param h5_path: HDF5 path for storage
    """
    test_study_names = ["Test1", "Test2", "Test3", "Test4"]
    pkdata = query_pkdb_data(h5_path=h5_path, study_names=test_study_names)

    # FIXME: https://github.com/matthiaskoenig/pkdb_analysis/issues/25
    # print(pkdata.studies)
    # assert len(pkdata.studies) == len(test_study_names)
    print(pkdata.interventions)
    pkdata.to_hdf5(h5_path)


if __name__ == "__main__":
    create_h5_test_data(h5_path=TEST_HDF5)



from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from pkdb_analysis.data import PKData
from pkdb_analysis.inference.body_weight import (
    infer_intervention,
    infer_output,
    infer_weight,
    ureg,
)
from pkdb_analysis.meta_analysis import MetaAnalysis
from pkdb_analysis.test import TEST_HDF5
from pkdb_analysis.test.data.group_data import (
    GROUP_OUTPUT,
    INDIVIDUAL_NO_BODYWEIGHT_OUTPUT,
    INDIVIDUAL_OUTPUT,
)


def test_output_inference_by_body_weight1():
    series = pd.Series(INDIVIDUAL_OUTPUT)
    output = infer_output(series)
    assert output["per_bw"] is True
    assert output["inferred"] is True
    assert output.value == 2 / 80
    assert output["mean"] is None
    assert ureg(output.unit) == ureg("ng/kg")
    assert output.intervention_value == 0.8
    assert ureg(output.intervention_unit) == ureg("ng")


def test_output_inference_by_body_weight2():
    series = pd.Series(GROUP_OUTPUT)
    output = infer_output(series)
    assert output["inferred"] is True
    assert output["per_bw"] is True
    assert output["mean"] == 4.5 / 80
    assert output.cv == 0.4
    assert ureg(output.unit) == ureg("ng/kg")
    assert np.isclose(output.intervention_value, 0.8)
    assert ureg(output.intervention_unit) == ureg("ng")


def test_output_inference_by_body_weight3():
    series = pd.Series(INDIVIDUAL_NO_BODYWEIGHT_OUTPUT)
    output = infer_output(series)
    assert output.isna().all()


def test_intervention_inference_by_body_weight1():
    series = pd.Series(INDIVIDUAL_OUTPUT)
    output = infer_intervention(series)
    assert output["inferred"] is True
    assert output["intervention_per_bw"] is True
    assert output.value == 2
    assert output["mean"] is None
    assert ureg(output.unit) == ureg("ng")
    assert np.isclose(output.intervention_value, 0.8 / 80)
    assert ureg(output.intervention_unit) == ureg("ng/kg")


def test_inference_by_body_weight1():
    df = pd.DataFrame([INDIVIDUAL_OUTPUT])
    outputs = infer_weight(df)
    assert len(outputs) == 4
    for group, output_subset in outputs.groupby(["intervention_per_bw", "per_bw"]):
        assert len(output_subset) == 1


def test_inference_by_body_weight2():
    df = pd.DataFrame([INDIVIDUAL_OUTPUT, GROUP_OUTPUT])
    outputs = infer_weight(df)
    assert len(outputs) == 8
    for group, output_subset in outputs.groupby(["intervention_per_bw", "per_bw"]):
        assert len(output_subset) == 2


# FIXME
@pytest.mark.skip("FIXME")
def test_inference_by_body_weight3():
    test_data = PKData.from_hdf5(TEST_HDF5)
    ma = MetaAnalysis(test_data, ["caffeine"], "test/url/")
    ma.create_results()
    results_inferred = infer_weight(ma.results)
    assert len(results_inferred) > len(ma.results)



import math
from itertools import combinations

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.cluster import AgglomerativeClustering
from sklearn.preprocessing import Normalizer


def _classify(df, n_clusters):
    coocurence_matrix = df.pivot_table(
        index="source", columns="target", aggfunc="sum"
    ).fillna(0)
    norm_coocurence_matrix = Normalizer(norm="l2").fit_transform(coocurence_matrix)
    norm_coocurence_matrix = pd.DataFrame(
        norm_coocurence_matrix,
        index=coocurence_matrix.index,
        columns=coocurence_matrix.columns,
    )
    classification_model = AgglomerativeClustering(
        linkage="single", n_clusters=n_clusters
    )
    classes = classification_model.fit_predict(norm_coocurence_matrix.T.corr())
    nodes = pd.DataFrame()
    nodes["name"] = norm_coocurence_matrix.index
    nodes["group"] = classes
    return nodes


def _detail_link(study):
    substances = [substance["name"] for substance in study.substances]
    substances = combinations(substances, 2)
    substances = pd.DataFrame(substances, columns=["source", "target"])
    substances["study_sid"] = study.sid
    substances["study_name"] = study.name
    return substances


def _basic_links(df):
    df["value"] = 1
    df2 = pd.DataFrame()
    df2["source"] = df["target"]
    df2["target"] = df["source"]
    df2["value"] = df["value"]
    df2["study_sid"] = df["study_sid"]
    df2["study_name"] = df["study_name"]
    return df2.append(df, ignore_index=True)


def _updated_links(basic_links, nodes):
    links = pd.DataFrame()
    map_s = pd.Series(nodes.name).reset_index().set_index("name")["index"]

    links["source"] = basic_links["source"].apply(lambda x: map_s.loc[x])
    links["target"] = basic_links["target"].apply(lambda x: map_s.loc[x])
    links["value"] = basic_links["value"]
    links["study_sid"] = basic_links["study_sid"]  # not jet used
    links["study_name"] = basic_links["study_name"]  # not jet used

    links = links.groupby(["source", "target"]).sum().reset_index()

    return links


def show_values_on_bars(axs, bottom):
    def _show_on_single_plot(ax):
        for p in ax.patches:
            _x = p.get_x() + p.get_width() / 2
            _y = p.get_y() + bottom

            try:
                value = int(p.get_height())
            except ValueError:
                continue

            ax.text(
                _x,
                _y,
                value,
                ha="center",
            )

    if isinstance(axs, np.ndarray):
        for idx, ax in np.ndenumerate(axs):
            _show_on_single_plot(ax)
    else:
        _show_on_single_plot(axs)


def arc_plot(pkdata, n_clusters=7):
    df = pd.DataFrame([], columns=["source", "target", "study"])
    for study in pkdata.studies.itertuples():
        df = df.append(_detail_link(study))
    basic_links = _basic_links(df)
    nodes = _classify(basic_links, n_clusters)
    links = _updated_links(basic_links, nodes)
    json_dict = {}
    json_dict["nodes"] = nodes.to_dict(orient="records")
    json_dict["links"] = links.to_dict(orient="records")
    return json_dict


def study_overview(pkdata, title=None, max_values={}, **kwargs):
    fig, axes = plt.subplots(ncols=1, nrows=5, sharex=True, **kwargs)
    if title is not None:
        fig.suptitle(title, fontsize=16)
    sns.set_style("white")
    sns.despine(left=True, bottom=True)
    df = pkdata.studies.df.set_index("name")[
        [
            "group_count",
            "individual_count",
            "intervention_count",
            "output_count",
            "output_calculated_count",
            "timecourse_count",
        ]
    ].replace({0, 0.1})

    colors = sns.color_palette("Set1", n_colors=8, desat=0.5)

    plot_info = [
        {"name": "groups", "col_name": "group_count", "color": colors[0], "ax": 0},
        {
            "name": "individuals",
            "col_name": "individual_count",
            "color": colors[1],
            "ax": 1,
        },
        {
            "name": "interventions",
            "col_name": "intervention_count",
            "color": colors[2],
            "ax": 2,
        },
        {"name": "outputs", "col_name": "output_count", "color": colors[3], "ax": 3},
        {
            "name": "timecourses",
            "col_name": "timecourse_count",
            "color": colors[4],
            "ax": 4,
        },
    ]

    plot_info = pd.DataFrame(plot_info, columns=["ax", "name", "col_name", "color"])
    color_timecourse = colors[4]
    color_output = colors[3]

    for subplot in plot_info.itertuples():
        ax = axes[subplot.ax]
        if subplot.name == "outputs":
            df["output_reported_count"] = (
                df["output_count"] - df["output_calculated_count"]
            )
            data = (
                df[["output_calculated_count", "output_reported_count"]]
                .stack()
                .reset_index(1)
                .rename(columns={"level_1": "type", 0: "value"})
                .replace(
                    {
                        "output_calculated_count": "Calculated",
                        "output_reported_count": "Reported",
                    }
                )
            )

            sns.barplot(
                data=data,
                x=data.index,
                y="value",
                hue=data.type,
                ax=ax,
                palette=[color_timecourse, color_output],
            )
            ax.get_legend().remove()

        else:
            sns.barplot(
                data=df, x=df.index, y=subplot.col_name, ax=ax, color=subplot.color
            )

        ax.set_yscale("log")
        param = {"bottom": 0.1}
        if max_values.get(subplot.name, False):
            param["top"] = max_values.get(subplot.name)
        ax.set_ylim(**param)

        ax.set_xlabel("")

        ax.set_ylabel(subplot.name, rotation=0, labelpad=5, horizontalalignment="right")
        ax.set_yticks([])
        show_values_on_bars(ax, 0.2)
    plt.setp(
        axes[4].get_xticklabels(), rotation=70, horizontalalignment="right", fontsize=8
    )
    return fig


def study_overview_wrapped(pkdata, rows):
    rows_per_plot = math.ceil(len(pkdata.studies) / rows)
    index = 0
    max_values = {
        "groups": pkdata.studies.group_count.max(),
        "individuals": pkdata.studies.individual_count.max(),
        "interventions": pkdata.studies.intervention_count.max(),
        "outputs": pkdata.studies.output_count.max(),
        "timecourses": pkdata.studies.timecourse_count.max(),
    }
    for row in range(rows):
        subset = pkdata.copy()
        new_index = index + rows_per_plot
        subset.studies = subset.studies.iloc[index:new_index]
        index = new_index
        subset._concise()
        fig = study_overview(
            subset, max_values=max_values, figsize=(len(subset.studies), 6)
        )
        fig.savefig(
            f"./caffeine_overview_{rows}-{row}.png", dpi=300, bbox_inches="tight"
        )
